""" write Component from YAML file

"""

from typing import Union, IO, Any
import pathlib
import io
from omegaconf import OmegaConf

from pp.component import Component
from pp.components import component_type2factory as component_type2factory_default
from pp.routing import link_optical_ports

valid_placements = ["x", "y", "rotation", "mirror"]
valid_keys = [
    "name",
    "instances",
    "placements",
    "connections",
    "ports",
    "routes",
    "bundle_routes",
]

sample_mmis = """
name:
    mmis

instances:
    mmi_long:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 10
    mmi_short:
      component: mmi1x2
      settings:
        width_mmi: 4.5
        length_mmi: 5

placements:
    mmi_long:
        rotation: 180
        x: 100
        y: 100

routes:
    mmi_short,E1: mmi_long,E0

ports:
    E0: mmi_short,W0
    W0: mmi_long,W0
"""


sample_connections = """
instances:
    wgw:
      component: waveguide
      settings:
        width: 1
        length: 1
    wgn:
      component: waveguide
      settings:
        width: 0.5
        length: 0.5

connections:
    wgw,E0: wgn,W0

"""


sample_mirror = """
name:
    mirror

instances:
    CP1:
      component: mmi1x2
      settings:
          width_mmi: 4.5
          length_mmi: 10
    CP2:
        component: mmi1x2
        settings:
            width_mmi: 4.5
            length_mmi: 5
    arm_top:
        component: mzi_arm
        settings:
            L0: 30
    arm_bot:
        component: mzi_arm
        settings:
            L0: 30

placements:
    arm_bot:
        mirror: [0, 0, 0, 10]
        rotation: 180
ports:
    W0: CP1,W0
    E0: CP2,W0

connections:
    arm_bot,W0: CP1,E0
    arm_top,W0: CP1,E1
    CP2,E0: arm_bot,E0
    CP2,E1: arm_top,E0
"""


def component_from_yaml(
    yaml: Union[str, pathlib.Path, IO[Any]], component_type2factory=None, **kwargs
) -> Component:
    """Returns a Component defined from YAML

    name: name of Component
    instances:
        name
        component
        settings
    placements: x, y and rotations
    connections: between instances
    ports (Optional): defines ports to expose
    routes (Optional): defines routes
    ports (Optional): defines ports to expose

    Args:
        yaml: YAML IO describing Component (instances, placements, routing, ports, connections)
        component_type2factory: dict of {factory_name: factory_function}
        kwargs: cache, pins ... to pass to all factories

    Returns:
        Component

    """
    yaml = io.StringIO(yaml) if isinstance(yaml, str) and "\n" in yaml else yaml
    component_type2factory = component_type2factory or component_type2factory_default

    conf = OmegaConf.load(yaml)
    for key in conf.keys():
        assert key in valid_keys, f"{key} not in {list(valid_keys)}"

    instances = {}
    routes = {}
    name = conf.get("name") or "Unnamed"
    c = Component(name)
    placements_conf = conf.get("placements")
    routes_conf = conf.get("routes")
    bundle_routes_conf = conf.get("bundle_routes")
    ports_conf = conf.get("ports")
    connections_conf = conf.get("connections")

    for instance_name in conf.instances:
        instance_conf = conf.instances[instance_name]
        component_type = instance_conf["component"]
        assert (
            component_type in component_type2factory
        ), f"{component_type} not in {list(component_type2factory.keys())}"
        component_settings = instance_conf["settings"] or {}
        component_settings.update(**kwargs)
        ci = component_type2factory[component_type](**component_settings)
        ci.name = instance_name
        ref = c << ci
        instances[instance_name] = ref

        if placements_conf:
            placement_settings = placements_conf[instance_name] or {}
            for k, v in placement_settings.items():
                if k not in valid_placements:
                    raise ValueError(
                        f"`{k}` not valid placement {valid_placements} for"
                        f" {instance_name}"
                    )
                elif k == "rotation":
                    ref.rotate(v, (ci.x, ci.y))
                elif k == "mirror":
                    ref.mirror((v[0], v[1]), (v[2], v[3]))
                else:
                    setattr(ref, k, v)

    if connections_conf:
        for port_src_string, port_dst_string in connections_conf.items():
            instance_src_name, port_src_name = port_src_string.split(",")
            instance_dst_name, port_dst_name = port_dst_string.split(",")

            instance_src_name = instance_src_name.strip()
            instance_dst_name = instance_dst_name.strip()
            port_src_name = port_src_name.strip()
            port_dst_name = port_dst_name.strip()

            assert (
                instance_src_name in instances
            ), f"{instance_src_name} not in {list(instances.keys())}"
            assert (
                instance_dst_name in instances
            ), f"{instance_dst_name} not in {list(instances.keys())}"
            instance_src = instances[instance_src_name]
            instance_dst = instances[instance_dst_name]

            assert port_src_name in instance_src.ports, (
                f"{port_src_name} not in {list(instance_src.ports.keys())} for"
                f" {instance_src_name} "
            )
            assert port_dst_name in instance_dst.ports, (
                f"{port_dst_name} not in {list(instance_dst.ports.keys())} for"
                f" {instance_dst_name}"
            )
            port_dst = instance_dst.ports[port_dst_name]
            instance_src.connect(port=port_src_name, destination=port_dst)

    if routes_conf:
        for port_src_string, port_dst_string in routes_conf.items():
            instance_src_name, port_src_name = port_src_string.split(",")
            instance_dst_name, port_dst_name = port_dst_string.split(",")

            instance_src_name = instance_src_name.strip()
            instance_dst_name = instance_dst_name.strip()
            port_src_name = port_src_name.strip()
            port_dst_name = port_dst_name.strip()

            assert (
                instance_src_name in instances
            ), f"{instance_src_name} not in {list(instances.keys())}"
            assert (
                instance_dst_name in instances
            ), f"{instance_dst_name} not in {list(instances.keys())}"

            instance_in = instances[instance_src_name]
            instance_out = instances[instance_dst_name]

            assert port_src_name in instance_in.ports, (
                f"{port_src_name} not in {list(instance_in.ports.keys())} for"
                f" {instance_src_name} "
            )
            assert port_dst_name in instance_out.ports, (
                f"{port_dst_name} not in {list(instance_out.ports.keys())} for"
                f" {instance_dst_name}"
            )

            port_src = instance_in.ports[port_src_name]
            port_out = instance_out.ports[port_dst_name]

            route = link_optical_ports([port_src], [port_out])
            c.add(route)
            routes[f"{port_src_string}:{port_dst_string}"] = route[0]

    if bundle_routes_conf:
        for route in bundle_routes_conf:
            ports1 = []
            ports2 = []
            routes = bundle_routes_conf[route]
            for port_src_string, port_dst_string in routes.items():
                instance_src_name, port_src_name = port_src_string.split(",")
                instance_dst_name, port_dst_name = port_dst_string.split(",")

                instance_src_name = instance_src_name.strip()
                instance_dst_name = instance_dst_name.strip()
                port_src_name = port_src_name.strip()
                port_dst_name = port_dst_name.strip()

                assert (
                    instance_src_name in instances
                ), f"{instance_src_name} not in {list(instances.keys())}"
                assert (
                    instance_dst_name in instances
                ), f"{instance_dst_name} not in {list(instances.keys())}"

                instance_in = instances[instance_src_name]
                instance_out = instances[instance_dst_name]

                assert port_src_name in instance_in.ports, (
                    f"{port_src_name} not in {list(instance_in.ports.keys())} for"
                    f" {instance_src_name} "
                )
                assert port_dst_name in instance_out.ports, (
                    f"{port_dst_name} not in {list(instance_out.ports.keys())} for"
                    f" {instance_dst_name}"
                )

                ports1.append(instance_in.ports[port_src_name])
                ports2.append(instance_out.ports[port_dst_name])
            route = link_optical_ports(ports1, ports2)
            c.add(route)

    if ports_conf:
        assert hasattr(ports_conf, "items"), f"{ports_conf} needs to be a dict"
        for port_name, instance_comma_port in ports_conf.items():
            instance_name, instance_port_name = instance_comma_port.split(",")
            instance_name = instance_name.strip()
            instance_port_name = instance_port_name.strip()
            assert (
                instance_name in instances
            ), f"{instance_name} not in {list(instances.keys())}"
            instance = instances[instance_name]
            assert instance_port_name in instance.ports, (
                f"{instance_port_name} not in {list(instance.ports.keys())} for"
                f" {instance_name} "
            )
            c.add_port(port_name, port=instance.ports[instance_port_name])
    c.instances = instances
    c.routes = routes
    return c


def test_sample():
    c = component_from_yaml(sample_mmis)
    assert len(c.get_dependencies()) == 3
    assert len(c.ports) == 2
    return c


def test_connections():
    c = component_from_yaml(sample_connections)
    # print(len(c.get_dependencies()))
    # print(len(c.ports))
    assert len(c.get_dependencies()) == 2
    assert len(c.ports) == 0
    return c


def test_mirror():
    c = component_from_yaml(sample_mirror)
    # print(len(c.get_dependencies()))
    # print(len(c.ports))
    assert len(c.get_dependencies()) == 3
    assert len(c.ports) == 2
    return c


sample_2x2_connections_problem = """
name:
    connections_2x2_problem

instances:
    mmi_bottom:
      component: mmi2x2
      settings:
            length_mmi: 5
    mmi_top:
      component: mmi2x2
      settings:
            length_mmi: 5

placements:
    mmi_top:
        x: 100
        y: 100

routes:
    mmi_bottom,E0: mmi_top,W0
    mmi_bottom,E1: mmi_top,W1

"""


def test_connections_2x2_problem():
    c = component_from_yaml(sample_2x2_connections_problem, pins=True, cache=False)
    # print(len(c.get_dependencies()))
    # print(len(c.ports))
    assert len(c.get_dependencies()) == 4
    assert len(c.ports) == 0
    return c


sample_2x2_connections_solution = """
name:
    connections_2x2_solution

instances:
    mmi_bottom:
      component: mmi2x2
      settings:
            length_mmi: 5
    mmi_top:
      component: mmi2x2
      settings:
            length_mmi: 5

placements:
    mmi_top:
        x: 100
        y: 100

bundle_routes:
    mmis:
        mmi_bottom,E0: mmi_top,W0
        mmi_bottom,E1: mmi_top,W1

"""


def test_connections_2x2_solution():
    c = component_from_yaml(sample_2x2_connections_solution, pins=True, cache=False)
    # print(len(c.get_dependencies()))
    # print(len(c.ports))
    assert len(c.get_dependencies()) == 4
    assert len(c.ports) == 0
    return c


if __name__ == "__main__":

    # c = test_connections_2x2_problem()
    # c = test_connections_2x2_solution()
    # pp.show(c)

    # test_sample()
    # test_connections()
    test_mirror()
    # c = test_mirror()

    # sample = sample_mmis
    # sample = sample_connections
    sample = sample_mirror
    c = component_from_yaml(sample)

    # c = component_from_yaml(sample_connections)
    # assert len(c.get_dependencies()) == 3
    # test_component_from_yaml()
    # test_component_from_yaml_with_routing()
    # print(c.ports)
    # c = pp.routing.add_fiber_array(c)



import numpy as np
import pp
from pp.sp.write import write


def plot(
    component_or_results_dict,
    logscale=True,
    keys=None,
    height_nm=220,
    dirpath=pp.CONFIG["sp"],
    **kwargs,
):
    """ plots Sparameters

    Args:
        component_or_results_dict:
        logscale: plots 20*log10(results)
        keys: list of keys to plot
        height_nm: nm height
        dirpath: where to store the simulations
        **kwargs:
            layer2nm: dict of {(1,0): 220}
            layer2material: dict of {(1,0): "Silicon ..."
            remove_layers: list of tuples (layers to remove)
            background_material: for the background
            port_width: port width (m)
            port_height: port height (m)
            port_extension_um: port extension (um)
            mesh_accuracy: 2 (1: coarse, 2: fine, 3: superfine)
            zmargin: for the FDTD region 1e-6 (m)
            ymargin: for the FDTD region 2e-6 (m)
            wavelength_start: 1.2e-6 (m)
            wavelength_stop: 1.6e-6 (m)
            wavelength_points: 500

    """
    import matplotlib.pyplot as plt

    r = component_or_results_dict
    if isinstance(r, pp.Component):
        r = write(component=r, height_nm=height_nm, dirpath=dirpath, **kwargs)
    w = r["wavelength_nm"]

    if keys:
        assert isinstance(keys, list)
        for key in keys:
            assert key in r, f"{key} not in {r.keys()}"
    else:
        keys = [key for key in r.keys() if key.startswith("S") and key.endswith("m")]

    for key in keys:
        if logscale:
            y = 20 * np.log10(r[key])
        else:
            y = r[key]

        plt.plot(w, y, label=key[:-1])
    plt.legend()
    plt.xlabel("wavelength (nm)")
    if logscale:
        plt.ylabel("Transmission (dB)")
    else:
        plt.ylabel("Transmission")


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    remove_layers = []
    layer2nm = {(1, 0): 220}

    # r = write(component=pp.c.waveguide(), layer2nm=layer2nm)
    # r = write(component=pp.c.mmi2x2(), layer2nm=layer2nm)
    # r = write(component=pp.c.mmi1x2(), layer2nm=layer2nm)
    r = write(component=pp.c.coupler(), layer2nm=layer2nm)
    # r = write(component=pp.c.bend_circular(), layer2nm=layer2nm)
    plot(r, logscale=True)
    plt.show()



"""
find GDS labels and write labels into a CSV file
"""

import pathlib
import csv
import klayout.db as pya

from pp import LAYER


def find_labels(gdspath, label_layer=LAYER.LABEL, prefix="opt_"):
    """ finds labels and locations from a GDS file """
    # Load the layout
    gdspath = str(gdspath)
    layout = pya.Layout()
    layout.read(gdspath)

    # Get the top cell and the units, and find out the index of the layer
    topcell = layout.top_cell()
    dbu = layout.dbu
    layer = pya.LayerInfo(label_layer[0], label_layer[1])
    layer_index = layout.layer(layer)

    # Extract locations
    iterator = topcell.begin_shapes_rec(layer_index)

    while not (iterator.at_end()):
        shape, trans = iterator.shape(), iterator.trans()
        iterator.next()
        if shape.is_text():
            text = shape.text
            if text.string.startswith(prefix):
                transformed = text.transformed(trans)
                yield text.string, transformed.x * dbu, transformed.y * dbu


def write_labels(gdspath, label_layer=LAYER.LABEL, csv_filename=None, prefix="opt_"):
    """Load  GDS mask and extracts the labels and coordinates from a GDS file"""
    labels = list(find_labels(gdspath, label_layer=label_layer, prefix=prefix))

    # Save the coordinates somewhere sensible
    if csv_filename is None:
        gdspath = pathlib.Path(gdspath)
        csv_filename = gdspath.with_suffix(".csv")
    with open(csv_filename, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerows(labels)
    print(f"Wrote {csv_filename}")


if __name__ == "__main__":
    from pp.config import CONFIG

    gdspath = CONFIG["samples_path"] / "mask" / "build" / "mask" / "sample.gds"
    write_labels(gdspath)



"""
Lets create a new component.

We create a function which returns a pp.Component instance.

Lets build waveguide crossing out of a vertical and horizonal arm


- Create a component instance that we will return from a component_factory function. We will use a autoname decorator to define the name
- Define a polygon
- Create ports
"""

from phidl import quickplot as qp
import pp
from pp import LAYER


@pp.autoname
def test_crossing_arm(wg_width=0.5, r1=3.0, r2=1.1, taper_width=1.2, taper_length=3.4):
    """ crossing arm
    """
    c = pp.Component()
    c << pp.c.ellipse(radii=(r1, r2), layer=LAYER.SLAB150)

    xmax = taper_length + taper_width / 2
    h = wg_width / 2
    taper_points = [
        (-xmax, h),
        (-taper_width / 2, taper_width / 2),
        (taper_width / 2, taper_width / 2),
        (xmax, h),
        (xmax, -h),
        (taper_width / 2, -taper_width / 2),
        (-taper_width / 2, -taper_width / 2),
        (-xmax, -h),
    ]

    c.add_polygon(taper_points, layer=LAYER.WG)

    c.add_port(
        name="W0", midpoint=(-xmax, 0), orientation=180, width=wg_width, layer=LAYER.WG
    )

    c.add_port(
        name="E0", midpoint=(xmax, 0), orientation=0, width=wg_width, layer=LAYER.WG
    )
    return c


if __name__ == "__main__":
    c = test_crossing_arm()
    pp.show(c)
    qp(c)



from typing import Tuple
import pp
from pp.component import Component
from pp.name import autoname
from pp.layers import LAYER


@autoname
def cross(
    length: float = 10.0,
    width: float = 3.0,
    layer: Tuple[int, int] = LAYER.WG,
    port_type="optical",
) -> Component:
    """Generates a right-angle cross from two rectangles of specified length and width.

    Args:
        length: float Length of the cross from one end to the other
        width: float Width of the arms of the cross
        layer: int, array-like[2], or set Specific layer(s) to put polygon geometry on

    .. plot::
      :include-source:

      import pp

      c = pp.c.cross(length=10, width=3, layer=(1, 0))
      pp.plotgds(c)

    """

    c = pp.Component()
    R = pp.c.rectangle(size=(width, length), layer=layer)
    r1 = c.add_ref(R).rotate(90)
    r2 = c.add_ref(R)
    r1.center = (0, 0)
    r2.center = (0, 0)
    c.add_port(
        "E0",
        width=width,
        layer=layer,
        port_type=port_type,
        orientation=0,
        midpoint=(+length / 2, 0),
    )
    c.add_port(
        "W0",
        width=width,
        layer=layer,
        port_type=port_type,
        orientation=180,
        midpoint=(-length / 2, 0),
    )
    c.add_port(
        "N0",
        width=width,
        layer=layer,
        port_type=port_type,
        orientation=90,
        midpoint=(0, length / 2),
    )
    c.add_port(
        "S0",
        width=width,
        layer=layer,
        port_type=port_type,
        orientation=270,
        midpoint=(0, -length / 2),
    )
    return c


if __name__ == "__main__":
    c = cross()
    cc = pp.routing.add_fiber_array(c)
    # print(c.ports)
    pp.show(cc)



from typing import Callable, Tuple, Optional
import numpy as np

import picwriter.components as pc

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp
from pp.component import Component


@pp.autoname
def cdc(
    length: float = 30.0,
    gap: float = 0.5,
    period: float = 0.220,
    dc: float = 0.5,
    angle: float = np.pi / 6.0,
    width_top: float = 2.0,
    width_bot: float = 0.75,
    input_bot: bool = False,
    dw_top: Optional[float] = None,
    dw_bot: Optional[float] = None,
    fins: bool = False,
    fin_size: Tuple[float, float] = (0.2, 0.05),
    contradc_wgt: None = None,
    port_midpoint: Tuple[int, int] = (0, 0),
    direction: str = "EAST",
    waveguide_template: Callable = wg_strip,
    **kwargs
) -> Component:
    """Grating-Assisted Contra-Directional Coupler

    Args:
       length (float): Length of the coupling region.
       gap (float): Distance between the two waveguides.
       period (float): Period of the grating.
       dc (float): Duty cycle of the grating. Must be between 0 and 1.
       angle (float): Angle in radians (between 0 and pi/2) at which the waveguide bends towards the coupling region.  Default=pi/6.
       width_top (float): Width of the top waveguide in the coupling region.  Defaults to the WaveguideTemplate wg width.
       width_bot (float): Width of the bottom waveguide in the coupling region.  Defaults to the WaveguideTemplate wg width.
       dw_top (float): Amplitude of the width variation on the top.  Default=gap/2.0.
       dw_bot (float): Amplitude of the width variation on the bottom.  Default=gap/2.0.
       input_bot (boolean): If `True`, will make the default input the bottom waveguide (rather than the top).  Default=`False`
       fins (boolean): If `True`, adds fins to the input/output waveguides.  In this case a different template for the component must be specified.  This feature is useful when performing electron-beam lithography and using different beam currents for fine features (helps to reduce stitching errors).  Defaults to `False`
       fin_size ((x,y) Tuple): Specifies the x- and y-size of the `fins`.  Defaults to 200 nm x 50 nm
       contradc_wgt (WaveguideTemplate): If `fins` above is True, a WaveguideTemplate (contradc_wgt) must be specified.  This defines the layertype / datatype of the ContraDC (which will be separate from the input/output waveguides).  Defaults to `None`
       port_midpoint (tuple): Cartesian coordinate of the input port (AT TOP if input_bot=False, AT BOTTOM if input_bot=True).  Defaults to (0,0).
       direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians).  Defaults to 'EAST'.
       waveguide_template (WaveguideTemplate): Picwriter WaveguideTemplate object


    .. plot::
      :include-source:

      import pp

      c = pp.c.cdc()
      pp.plotgds(c)

    """

    c = pc.ContraDirectionalCoupler(
        pp.call_if_func(wg_strip, **kwargs),
        length=length,
        gap=gap,
        period=period,
        dc=dc,
        angle=angle,
        width_top=width_top,
        width_bot=width_bot,
        dw_top=dw_top,
        dw_bot=dw_bot,
        input_bot=input_bot,
        fins=fins,
        fin_size=fin_size,
        contradc_wgt=contradc_wgt,
        port=port_midpoint,
        direction=direction,
    )

    component = picwriter2component(c)
    pp.port.rename_ports_by_orientation(component)
    return component


if __name__ == "__main__":
    import pp

    c = cdc()
    print(c.ports.keys())
    pp.show(c)



# NOTE: import order matters. Only change the order if you know what you are doing
# level 0 components
from pp.components.waveguide import waveguide
from pp.components.waveguide_heater import waveguide_heater
from pp.components.waveguide_heater import wg_heater_connected
from pp.components.waveguide_pin import waveguide_pin

from pp.components.bend_circular import bend_circular
from pp.components.bend_circular import bend_circular180
from pp.components.bend_circular_heater import bend_circular_heater
from pp.components.bend_s import bend_s
from pp.components.bezier import bezier
from pp.components.euler.bend_euler import bend_euler90
from pp.components.euler.bend_euler import bend_euler180

from pp.components.coupler90 import coupler90
from pp.components.coupler_straight import coupler_straight
from pp.components.coupler_symmetric import coupler_symmetric
from pp.components.coupler_asymmetric import coupler_asymmetric
from pp.components.hline import hline

# basic shapes
from pp.components.circle import circle
from pp.components.compass import compass
from pp.components.cross import cross
from pp.components.crossing_waveguide import crossing
from pp.components.crossing_waveguide import crossing45
from pp.components.crossing_waveguide import compensation_path
from pp.components.ellipse import ellipse
from pp.components.label import label
from pp.components.rectangle import rectangle
from pp.components.rectangle import rectangle_centered
from pp.components.ring import ring
from pp.components.taper import taper
from pp.components.taper import taper_strip_to_ridge
from pp.components.text import text

# optical test structures
from pp.components.pcm.litho_calipers import litho_calipers
from pp.components.pcm.litho_star import litho_star
from pp.components.pcm.litho_steps import litho_steps
from pp.components.pcm.verniers import verniers

from pp.components.grating_coupler.elliptical import grating_coupler_elliptical_te
from pp.components.grating_coupler.elliptical import grating_coupler_elliptical_tm
from pp.components.grating_coupler.elliptical2 import grating_coupler_elliptical2
from pp.components.grating_coupler.uniform import grating_coupler_uniform
from pp.components.grating_coupler.grating_coupler_tree import grating_coupler_tree
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_tm
from pp.components.delay_snake import delay_snake
from pp.components.spiral import spiral
from pp.components.spiral_inner_io import spiral_inner_io_euler
from pp.components.spiral_inner_io import spiral_inner_io
from pp.components.spiral_external_io import spiral_external_io
from pp.components.spiral_circular import spiral_circular
from pp.components.cdc import cdc
from pp.components.dbr import dbr
from pp.components.dbr2 import dbr2

# electrical
from pp.components.electrical.wire import wire
from pp.components.electrical.wire import corner
from pp.components.electrical.pad import pad
from pp.components.electrical.pad import pad_array
from pp.components.electrical.tlm import via
from pp.components.electrical.tlm import via1
from pp.components.electrical.tlm import via2
from pp.components.electrical.tlm import via3
from pp.components.electrical.tlm import tlm
from pp.components.electrical.pads_shorted import pads_shorted

# electrical PCM
from pp.components.pcm.test_resistance import test_resistance
from pp.components.pcm.test_via import test_via

# level 1 components
from pp.components.cavity import cavity
from pp.components.coupler import coupler
from pp.components.coupler_ring import coupler_ring
from pp.components.coupler_adiabatic import coupler_adiabatic
from pp.components.coupler_full import coupler_full
from pp.components.disk import disk
from pp.components.ring_single import ring_single
from pp.components.ring_double import ring_double
from pp.components.ring_single_bus import ring_single_bus
from pp.components.ring_double_bus import ring_double_bus
from pp.components.mmi1x2 import mmi1x2
from pp.components.mmi2x2 import mmi2x2
from pp.components.mzi2x2 import mzi_arm
from pp.components.mzi2x2 import mzi2x2
from pp.components.mzi1x2 import mzi1x2
from pp.components.mzi import mzi
from pp.components.loop_mirror import loop_mirror

# level 2 components
from pp.components.component_lattice import component_lattice
from pp.components.component_sequence import component_sequence
from pp.components.splitter_tree import splitter_tree
from pp.components.splitter_chain import splitter_chain


# we will test each factory component hash, ports and properties """
component_type2factory = dict(
    bend_circular180=bend_circular180,
    bend_circular=bend_circular,
    bend_circular_heater=bend_circular_heater,
    bend_euler180=bend_euler180,
    bend_euler90=bend_euler90,
    bend_s=bend_s,
    bezier=bezier,
    cavity=cavity,
    cdc=cdc,
    circle=circle,
    compass=compass,
    compensation_path=compensation_path,
    component_lattice=component_lattice,
    component_sequence=component_sequence,
    corner=corner,
    coupler90=coupler90,
    coupler=coupler,
    coupler_adiabatic=coupler_adiabatic,
    coupler_asymmetric=coupler_asymmetric,
    coupler_full=coupler_full,
    coupler_ring=coupler_ring,
    coupler_straight=coupler_straight,
    coupler_symmetric=coupler_symmetric,
    cross=cross,
    crossing45=crossing45,
    crossing=crossing,
    dbr2=dbr2,
    dbr=dbr,
    delay_snake=delay_snake,
    disk=disk,
    ellipse=ellipse,
    grating_coupler_elliptical2=grating_coupler_elliptical2,
    grating_coupler_elliptical_te=grating_coupler_elliptical_te,
    grating_coupler_elliptical_tm=grating_coupler_elliptical_tm,
    grating_coupler_te=grating_coupler_te,
    grating_coupler_tm=grating_coupler_tm,
    grating_coupler_tree=grating_coupler_tree,
    grating_coupler_uniform=grating_coupler_uniform,
    hline=hline,
    label=label,
    litho_calipers=litho_calipers,
    litho_star=litho_star,
    litho_steps=litho_steps,
    loop_mirror=loop_mirror,
    mmi1x2=mmi1x2,
    mmi2x2=mmi2x2,
    mzi1x2=mzi1x2,
    mzi2x2=mzi2x2,
    mzi=mzi,
    mzi_arm=mzi_arm,
    pad=pad,
    pad_array=pad_array,
    pads_shorted=pads_shorted,
    rectangle=rectangle,
    rectangle_centered=rectangle_centered,
    ring=ring,
    ring_double=ring_double,
    ring_double_bus=ring_double_bus,
    ring_single=ring_single,
    ring_single_bus=ring_single_bus,
    spiral=spiral,
    spiral_circular=spiral_circular,
    spiral_external_io=spiral_external_io,
    spiral_inner_io=spiral_inner_io,
    spiral_inner_io_euler=spiral_inner_io_euler,
    splitter_chain=splitter_chain,
    splitter_tree=splitter_tree,
    taper=taper,
    taper_strip_to_ridge=taper_strip_to_ridge,
    test_resistance=test_resistance,
    test_via=test_via,
    text=text,
    tlm=tlm,
    verniers=verniers,
    via1=via1,
    via2=via2,
    via3=via3,
    via=via,
    waveguide=waveguide,
    waveguide_heater=waveguide_heater,
    waveguide_pin=waveguide_pin,
    wg_heater_connected=wg_heater_connected,
    wire=wire,
)


def component_factory(component_type, **settings):
    """ returns a component with settings """
    import pp

    if isinstance(component_type, pp.Component):
        return component_type
    elif callable(component_type):
        return component_type(**settings)
    elif component_type not in component_type2factory.keys():
        raise ValueError(
            "component types available: \n {}".format(
                "\n".join(component_type2factory.keys())
            )
        )
    return component_type2factory[component_type](**settings)


# all this components are available as pp.c
__all__ = list(component_type2factory.keys())
leaf_components = ["bend_circular", "bend_euler90", "coupler", "mmi1x2", "mmi2x2"]
_containers = set(["cavity"])
_skip_test = set(
    ["label", "text", "spiral_circular", "component_sequence", "tlm", "coupler"]
)
_components = set(__all__) - _containers - _skip_test



import picwriter.components as pc
from pp.port import auto_rename_ports

from pp.components.waveguide_template import wg_strip
from pp.picwriter2component import picwriter2component
import pp
from pp.component import Component
from typing import Callable, Tuple


@pp.autoname
def coupler_adiabatic(
    length1: float = 20.0,
    length2: float = 50.0,
    length3: float = 30.0,
    wg_sep: float = 1.0,
    input_wg_sep: float = 3.0,
    output_wg_sep: float = 3.0,
    dw: float = 0.1,
    port: Tuple[int, int] = (0, 0),
    direction: str = "EAST",
    waveguide_template: Callable = wg_strip,
    **kwargs
) -> Component:
    """ 50/50 adiabatic coupler
    Adiabatic Coupler Cell class.  Design based on asymmetric adiabatic 3dB coupler designs, such as those from https://doi.org/10.1364/CLEO.2010.CThAA2, https://doi.org/10.1364/CLEO_SI.2017.SF1I.5, and https://doi.org/10.1364/CLEO_SI.2018.STh4B.4.  Uses Bezier curves for the input, with poles set to half of the x-length of the S-bend.

    In this design, Region I is the first half of the input S-bend waveguide where the input waveguides widths taper by +dw and -dw, Region II is the second half of the S-bend waveguide with constant, unbalanced widths, Region III is the region where the two asymmetric waveguides gradually come together, Region IV is the coupling region where the waveguides taper back to the original width at a fixed distance from one another, and Region IV is the  output S-bend waveguide.

    Args:
        length1 (float): Length of the region that gradually brings the two assymetric waveguides together.  In this region the waveguide widths gradually change to be different by `dw`.
        length2 (float): Length of the coupling region, where the asymmetric waveguides gradually become the same width.
        length3 (float): Length of the output region where the two waveguides separate.
        wg_sep (float): Distance between the two waveguides, center-to-center, in the coupling region (Region 2).
        input_wg_sep (float): Separation of the two waveguides at the input, center-to-center.
        output_wg_sep (float): Separation of the two waveguides at the output, center-to-center.
        dw (float): Change in waveguide width.  In Region 1, the top arm tapers to the waveguide width+dw/2.0, bottom taper to width-dw/2.0.
        port (tuple): Cartesian coordinate of the input port (top left).  Defaults to (0,0).
        direction (string): Direction that the component will point *towards*, can be of type `'NORTH'`, `'WEST'`, `'SOUTH'`, `'EAST'`, OR an angle (float, in radians).  Defaults to 'EAST'.
        waveguide_template: object or function

    Other Parameters:
       wg_width: 0.5
       wg_layer: pp.LAYER.WG[0]
       wg_datatype: pp.LAYER.WG[1]
       clad_layer: pp.LAYER.WGCLAD[0]
       clad_datatype: pp.LAYER.WGCLAD[1]
       bend_radius: 10
       cladding_offset: 3

    """

    c = pc.AdiabaticCoupler(
        pp.call_if_func(waveguide_template, **kwargs),
        length1=length1,
        length2=length2,
        length3=length3,
        wg_sep=wg_sep,
        input_wg_sep=input_wg_sep,
        output_wg_sep=output_wg_sep,
        dw=dw,
        port=port,
        direction=direction,
    )

    c = picwriter2component(c)
    c = auto_rename_ports(c)
    return c


if __name__ == "__main__":
    import pp

    c = coupler_adiabatic(length3=5)
    print(c.ports)
    pp.show(c)



from typing import Any, Callable, List, Optional, Tuple, Union
from numpy import float64
from phidl.device_layout import Label
from pp.component import Component, ComponentReference

from pp.routing.utils import direction_ports_from_list_ports

from pp.routing.connect import connect_elec_waypoints
from pp.routing.connect import get_waypoints_connect_strip
from pp.port import select_electrical_ports
from pp.components.electrical.pad import pad
import pp


def route_pad_array(
    component: Component,
    pad_spacing: float = 150.0,
    pad: Callable = pad,
    fanout_length: Optional[int] = None,
    max_y0_optical: None = None,
    waveguide_separation: float = 4.0,
    bend_radius: float = 0.1,
    connected_port_list_ids: None = None,
    n_ports: int = 1,
    excluded_ports: List[Any] = [],
    pad_indices: None = None,
    route_filter: Callable = connect_elec_waypoints,
    port_name: str = "W",
    pad_rotation: int = -90,
    x_pad_offset: int = 0,
    port_labels: None = None,
    select_ports: Callable = select_electrical_ports,
) -> Tuple[
    List[Union[ComponentReference, Label]], List[List[ComponentReference]], float64
]:
    """
    Returns component I/O elements for adding pads to south
    Basic routing, typically fine for small components
    No heuristic to avoid collisions between connectors.

    If specified ports to connect in a specific order
    (i.e if connected_port_list_ids is not None and not empty)
    then grab these ports

    Args:
        component: The component to connect.
        pad_spacing: the wanted spacing between the  pads
        fanout_length: Wanted distance between the gratings and the southest component port. If set to None, automatically calculated.
        max_y0_optical: Maximum y coordinate at which the intermediate optical ports can be set. Usually fine to leave at None.
        waveguide_separation: min separation between the waveguides used to route pads to the component I/O.
        bend_radius: bend radius
        list_port_labels: list of the port indices (e.g [0,3]) which require a T&M label.
        connected_port_list_ids: only for type 0 optical routing.  Can specify which ports goes to which pads assuming the gratings are ordered from left to right.  e.g ['N0', 'W1','W0','E0','E1', 'N1' ] or [4,1,7,3]
        n_ports: number of lines with I/O pads.  One line by default.  WARNING: Only works properly if:
            - n_ports divides the total number of ports
            - the components have an equal number of inputs and outputs
        pad_indices: allows to fine skip some grating slots e.g [0,1,4,5] will put two gratings separated by the pitch. Then there will be two empty pads slots, and after that an additional two gratings.

    Returns:
        elements, pads, y0_optical
    """
    if port_labels is None:
        ports = list(select_ports(component.ports).values())
    else:
        ports = [component.ports[lbl] for lbl in port_labels]

    ports = [p for p in ports if p.name not in excluded_ports]
    N = len(ports)
    if N == 0:
        return [], [], 0

    elements = []
    pad = pad() if callable(pad) else pad
    pads = [pad] * N

    route_filter_params = {
        "bend_radius": bend_radius,
        "wg_width": ports[0].width,
    }

    def routing_method(p1, p2, **kwargs):
        way_points = get_waypoints_connect_strip(p1, p2, **kwargs)
        return route_filter(way_points, **route_filter_params)

    """
    # `delta_gr_min` Used to avoid crossing between waveguides in special cases
    # This could happen when abs(x_port - x_grating) <= 2 * bend_radius
    """

    delta_gr_min = 2 * bend_radius + 1

    io_sep = pad_spacing
    offset = (N - 1) * io_sep / 2.0

    # Get the center along x axis
    x_c = round(sum([p.x for p in ports]) / N, 1)
    y_min = component.ymin  # min([p.y for p in ports])

    # Sort the list of optical ports:
    direction_ports = direction_ports_from_list_ports(ports)
    sep = waveguide_separation

    K = len(ports)
    K = K + 1 if K % 2 else K

    # Set routing type if not specified
    pxs = [p.x for p in ports]
    is_big_component = (
        (K > 2)
        or (max(pxs) - min(pxs) > io_sep - delta_gr_min)
        or (component.xsize > io_sep)
    )
    """
    Look at a bunch of conditions to choose the default length
    if the default fanout distance is not set
    """

    def has_p(side):
        return len(direction_ports[side]) > 0

    list_ew_ports_on_sides = [has_p(side) for side in ["E", "W"]]
    list_ns_ports_on_sides = [has_p(side) for side in ["N", "S"]]

    has_ew_ports = any(list_ew_ports_on_sides)
    has_ns_ports = any(list_ns_ports_on_sides)

    is_one_sided_horizontal = False
    for side1, side2 in [("E", "W"), ("W", "E")]:
        if len(direction_ports[side1]) >= 2:
            if all([len(direction_ports[side]) == 0 for side in ["N", "S", side2]]):
                is_one_sided_horizontal = True

    # Compute fanout length if not specified
    if fanout_length is None:
        fanout_length = bend_radius + 1.0
        # We need 3 bends in that case to connect the most bottom port to the
        if has_ew_ports and is_big_component:
            fanout_length = max(fanout_length, 3 * bend_radius + 1.0)

        if has_ns_ports or is_one_sided_horizontal:
            fanout_length = max(fanout_length, 2 * bend_radius + 1.0)

        if has_ew_ports and not is_big_component:
            fanout_length = max(fanout_length, bend_radius + 1.0)

    fanout_length += 5
    # use x for pad since we rotate it
    y0_optical = y_min - fanout_length - pad.ports[port_name].x
    y0_optical += -K / 2 * sep
    y0_optical = round(y0_optical, 1)

    if max_y0_optical is not None:
        y0_optical = round(min(max_y0_optical, y0_optical), 1)
    """
     - First connect half of the north ports going from middle of list
    down to first elements
     - then connect west ports (top to bottom)
     - then connect south ports (left to right)
     - then east ports (bottom to top)
     - then second half of the north ports (right to left)

    """
    north_ports = direction_ports["N"]
    north_start = north_ports[0 : len(north_ports) // 2]
    north_finish = north_ports[len(north_ports) // 2 :]

    west_ports = direction_ports["W"]
    west_ports.reverse()
    east_ports = direction_ports["E"]
    south_ports = direction_ports["S"]
    north_finish.reverse()  # Sort right to left
    north_start.reverse()  # Sort right to left
    ordered_ports = north_start + west_ports + south_ports + east_ports + north_finish

    nb_ports_per_line = N // n_ports
    pad_size_info = pad.size_info
    y_gap = (K / (n_ports) + 1) * sep
    y_sep = pad_size_info.height + y_gap + bend_radius

    offset = (nb_ports_per_line - 1) * io_sep / 2 - x_pad_offset
    io_pad_lines = []  # [[gr11, gr12, gr13...], [gr21, gr22, gr23...] ...]

    if pad_indices is None:
        pad_indices = list(range(nb_ports_per_line))
    else:
        assert len(pad_indices) == nb_ports_per_line

    for j in range(n_ports):
        pads = [
            pad.ref(
                position=(x_c - offset + i * io_sep, y0_optical - j * y_sep),
                rotation=pad_rotation,
                port_id=port_name,
            )
            for i, pad in zip(pad_indices, pads)
        ]

    io_pad_lines += [pads[:]]

    if connected_port_list_ids:
        ordered_ports = [component.ports[i] for i in connected_port_list_ids]

    for pads in io_pad_lines:
        for i in range(N):
            p0 = pads[i].ports[port_name]
            p1 = ordered_ports[i]
            elements += [routing_method(p0, p1, bend_radius=bend_radius)]

    return elements, io_pad_lines, y0_optical


if __name__ == "__main__":
    from pp.rotate import rotate

    c = pp.c.mzi2x2(with_elec_connections=True)
    c = rotate(c, 180)

    elements, pads, _ = route_pad_array(c, fanout_length=100)
    for e in elements:
        c.add(e)
    for e in pads:
        c.add(e)
    pp.show(c)



"""
updates GDS view in Klayout dynamically.
Requires the Klayout plugin installed in Klayout.
This happens when you run `bash install.sh` from the top of the gdsfactory package
"""

import os
import socket
import json
from pathlib import PosixPath


def show(gds_filename: PosixPath, keep_position: bool = True) -> None:
    """ Show GDS in klayout """
    if not os.path.isfile(gds_filename):
        raise ValueError("{} does not exist".format(gds_filename))
    data = {
        "gds": os.path.abspath(gds_filename),
        "keep_position": keep_position,
    }
    data = json.dumps(data)
    try:
        conn = socket.create_connection(("127.0.0.1", 8082), timeout=0.5)
        data = data + "\n"
        data = data.encode() if hasattr(data, "encode") else data
        conn.sendall(data)
        conn.close()
    except socket.error:
        print("warning, could not connect to the klive server")
        pass



from typing import List, Tuple
import pp
from pp.component import Component


@pp.autoname
def coupler_straight(
    length: float = 10.0,
    width: float = 0.5,
    gap: float = 0.27,
    layer: Tuple[int, int] = pp.LAYER.WG,
    layers_cladding: List[Tuple[int, int]] = [pp.LAYER.WGCLAD],
    cladding_offset: float = 3.0,
) -> Component:
    """ straight coupled waveguides. Two multimode ports

    .. plot::
      :include-source:

      import pp

      c = pp.c.coupler_straight()
      pp.plotgds(c)

    """

    c = Component()

    # Top path
    c.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=layer)
    y = width + gap

    # Bottom path
    c.add_polygon(
        [(0, y), (length, y), (length, width + y), (0, width + y)], layer=layer
    )

    # One multimode port on each side
    port_w = width * 2 + gap

    c.add_port(name="W0", midpoint=[0, port_w / 2], width=port_w, orientation=180)
    c.add_port(name="E0", midpoint=[length, port_w / 2], width=port_w, orientation=0)

    c.width = width
    c.length = length

    # cladding
    ymax = 2 * width + gap + cladding_offset
    for layer_cladding in layers_cladding:
        c.add_polygon(
            [
                (0, -cladding_offset),
                (length, -cladding_offset),
                (length, ymax),
                (0, ymax),
            ],
            layer=layer_cladding,
        )

    return c


@pp.autoname
def coupler_straight_biased(length=10, width=0.5, gap=0.27, layer=pp.LAYER.WG):
    return coupler_straight(
        width=pp.bias.width(width), gap=pp.bias.gap(gap), length=length, layer=layer
    )


def _demo():
    c = coupler_straight(gap=0.2)
    pp.write_gds(c)
    return c


if __name__ == "__main__":
    # c = _demo()
    c = coupler_straight_biased(width=0.5, gap=0.2)
    pp.show(c)



from pp.component import Component
from pp.name import autoname
from pp.layers import LAYER
from pp.components.rectangle import rectangle_centered


@autoname
def pads_shorted(width=100, n_pads=8, pad_spacing=150, layer=LAYER.M1):
    c = Component(name="shorted_pads")
    pad = rectangle_centered(x=width, y=width, layer=layer)
    for i in range(n_pads):
        pad_ref = c.add_ref(pad)
        pad_ref.movex(i * pad_spacing - n_pads / 2 * pad_spacing + pad_spacing / 2)

    short = rectangle_centered(x=pad_spacing * (n_pads - 1), y=10, layer=layer)
    c.add_ref(short)
    return c


if __name__ == "__main__":
    import pp

    c = pads_shorted()
    pp.show(c)



import pathlib
import configparser
import shutil

home = pathlib.Path.home()
config_path = home / ".gitconfig"

cwd = pathlib.Path.cwd()
src = cwd / ".gitattributes"
shutil.copy(src, home)

config = configparser.RawConfigParser()
config.read(config_path)
key = 'diff "gds_diff"'

if key not in config.sections():
    config.add_section(key)
    config.set(key, "command", "python -m gdsdiff.gds_diff_git")
    config.set(key, "binary", "True")

    with open(config_path, "w") as f:
        print("installing gds_diff")
        config.write(f, space_around_delimiters=True)



import uuid
from pp.component import Component


def connect_electrical_shortest_path(port1, port2):
    """connects two ports with a polygon that takes the shortest path"""
    c = Component(name="zz_conn_{}".format(uuid.uuid4()))
    layer = port1.layer
    p1x0 = port1.endpoints[0][0]
    p1y0 = port1.endpoints[0][1]
    p1x1 = port1.endpoints[1][0]
    p1y1 = port1.endpoints[1][1]

    p2x0 = port2.endpoints[0][0]
    p2y0 = port2.endpoints[0][1]
    p2x1 = port2.endpoints[1][0]
    p2y1 = port2.endpoints[1][1]

    if port1.orientation in [90, 270]:
        c.add_polygon(
            ([(p1x1, p1y0), (p1x0, p1y1), (p2x1, p2y1), (p2x0, p2y0)]), layer=layer
        )
    else:
        c.add_polygon(
            ([(p1x0, p1y1), (p1x1, p1y0), (p2x1, p2y1), (p2x0, p2y0)]), layer=layer
        )
    return c.ref()


if __name__ == "__main__":
    import pp
    from pp.components.electrical.pad import pad_array

    c = Component("mzi_with_pads")
    mzi = pp.c.mzi2x2(with_elec_connections=True)
    pads = pad_array(n=3, port_list=["S"])
    p = c << pads
    c << mzi
    p.move((-150, 200))
    ports_pads = list(p.ports.values())
    ports_mzi = mzi.get_ports_list(port_type="dc")

    for p1, p2 in zip(ports_pads, ports_mzi):
        c.add(connect_electrical_shortest_path(p1, p2))
    pp.show(c)



import pp


def test_name_from_args():
    name = "my_name"
    c = pp.Component(name)
    print(c)
    assert c.name == name


def test_name_in_kwargs():
    name = "my_name"
    c = pp.Component(name=name)
    print(c)
    assert c.name == name


def test_autoname():
    c = pp.c.waveguide(length=11)
    print(c)
    assert c.name == "waveguide_L11"


if __name__ == "__main__":
    test_autoname()
    # test_name_from_args()
    # test_name_in_kwargs()



""" one problem is that when we add references we have to make sure they have unique names

The photonics package `pp` has a decorator that names the objects that it produces depending on the parameters that we pass them

"""


import pp


@pp.autoname
def waveguide_autoname(width=10, height=1):
    wg = pp.Component("waveguide")
    wg.add_polygon([(0, 0), (width, 0), (width, height), (0, height)])
    wg.add_port(name="wgport1", midpoint=[0, height / 2], width=height, orientation=180)
    wg.add_port(
        name="wgport2", midpoint=[width, height / 2], width=height, orientation=0
    )
    return wg


c = waveguide_autoname()
print(c)

c = waveguide_autoname(width=0.5)
print(c)



import pytest


@pytest.fixture(scope="function")
def test_netlist_read():
    import pp

    c = pp.component_from_yaml(pp.CONFIG["netlists"] / "mzi.yml")
    # print(c.get_netlist().pretty())
    # print((c.get_netlist().connections.pretty()))
    # print(len(c.get_netlist().connections))
    # print(len(c.get_dependencies()))
    assert len(c.get_dependencies()) == 5
    assert len(c.get_netlist().connections) == 18


if __name__ == "__main__":
    c = test_netlist_read()



from typing import Tuple
import pp
from pp.component import Component


@pp.autoname
def compass(
    size: Tuple[float, float] = (4, 2), layer: Tuple[int, int] = pp.LAYER.WG,
) -> Component:
    """ Rectangular contact pad with centered ports on rectangle edges
    (north, south, east, and west)

    Args:
        size: tuple (4, 2)
        layer:

    .. plot::
      :include-source:

      import pp

      c = pp.c.compass(size=(4, 2), layer=1)
      pp.plotgds(c)
    """

    c = pp.Component()
    r = pp.c.rectangle(size=size, layer=layer)
    dx = size[0]
    dy = size[1]

    rr = r.ref(position=(-dx / 2, -dy / 2))
    c.add(rr)
    c.add_port(name="N", midpoint=[0, dy / 2], width=dx, orientation=90, layer=layer)
    c.add_port(name="S", midpoint=[0, -dy / 2], width=dx, orientation=-90, layer=layer)
    c.add_port(name="E", midpoint=[dx / 2, 0], width=dy, orientation=0, layer=layer)
    c.add_port(name="W", midpoint=[-dx / 2, 0], width=dy, orientation=180, layer=layer)

    return c


if __name__ == "__main__":
    c = compass(size=(1, 2), layer=pp.LAYER.WG)
    pp.show(c)



""" based on phidl tutorial

# Manipulating geometry 1 - Basic movement and rotation

There are several actions we can take to move and rotate the geometry.  These
actions include movement, rotation, and reflection.

"""


import pp

c = pp.Component()


wg1 = c << pp.c.waveguide(length=10, width=1)
wg2 = c << pp.c.waveguide(length=10, width=2, layer=pp.LAYER.SLAB90)


# wg2.move([10, 1])  # Shift the second waveguide we created over by dx = 10, dy = 4

# You can unconmment and play with the following move commands

# wg2.rotate(45) # Rotate waveguide by 45 degrees around (0,0)
# wg2.rotate(45, center=[5, 0])  # Rotate waveguide by 45 degrees around (5, 0)
# wg2.reflect(p1=[1, 1], p2=[1, 3])  # Reflects wg across the line formed by p1 and p2


pp.show(c)



""" add markers to Devices:

- pins
- outline

"""
import numpy as np
from pp.layers import LAYER, port_type2layer
from pp.port import read_port_markers
import pp


def _rotate(v, m):
    return np.dot(m, v)


def add_pin_triangle(component, port, layer=LAYER.PORT, label_layer=LAYER.TEXT):
    """
    # The port visualization pattern is a triangle with a right angle
    # The face opposite the right angle is the port width
    """
    p = port

    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dtip = np.array([d, 0])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptip = p.position + _rotate(dtip, rot_mat)
    polygon = [p0, p1, ptip]

    component.add_label(
        text=p.name, position=p.midpoint, layer=label_layer,
    )

    component.add_polygon(polygon, layer=layer)


def add_pin_square_inside(
    component, port, port_length=0.1, layer=LAYER.PORT, label_layer=LAYER.TEXT
):
    """ square towards the inside of the port

    .. code::
           _______________
          |               |
          |               |
          |               |
          ||              |
          ||              |
          |               |
          |      __       |
          |_______________|


    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2
    dx = port_length

    dbot = np.array([0, -d])
    dtop = np.array([0, d])
    dbotin = np.array([-dx, -d])
    dtopin = np.array([-dx, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)


def add_pin_square(
    component, port, port_length=0.1, layer=LAYER.PORT, label_layer=LAYER.PORT
):
    """ half out

    .. code::
           _______________
          |               |
          |               |
          |               |
         |||              |
         |||              |
          |               |
          |      __       |
          |_______________|
                 __

    """
    p = port
    a = p.orientation
    ca = np.cos(a * np.pi / 180)
    sa = np.sin(a * np.pi / 180)
    rot_mat = np.array([[ca, -sa], [sa, ca]])

    d = p.width / 2
    dx = port_length

    dbot = np.array([dx / 2, -d])
    dtop = np.array([dx / 2, d])
    dbotin = np.array([-dx / 2, -d])
    dtopin = np.array([-dx / 2, +d])

    p0 = p.position + _rotate(dbot, rot_mat)
    p1 = p.position + _rotate(dtop, rot_mat)
    ptopin = p.position + _rotate(dtopin, rot_mat)
    pbotin = p.position + _rotate(dbotin, rot_mat)
    polygon = [p0, p1, ptopin, pbotin]
    component.add_polygon(polygon, layer=layer)

    component.add_label(
        text=str(p.name), position=p.midpoint, layer=label_layer,
    )


def add_outline(component, layer=LAYER.DEVREC):
    """ adds devices outline in layer """
    c = component
    points = [
        [c.xmin, c.ymin],
        [c.xmax, c.ymin],
        [c.xmax, c.ymax],
        [c.xmin, c.ymax],
    ]
    c.add_polygon(points, layer=layer)


def add_pins(
    component,
    add_port_marker_function=add_pin_square,
    port_type2layer=port_type2layer,
    **kwargs,
):
    """ add port markers:

    Args:
        component: to add ports
        add_port_marker_function:
        port_type2layer: dict mapping port types to marker layers for ports

    Add device recognition layer
    """

    if hasattr(component, "ports") and component.ports:
        for p in component.ports.values():
            layer = port_type2layer[p.port_type]
            add_port_marker_function(
                component=component, port=p, layer=layer, label_layer=layer, **kwargs
            )


def add_pins_and_outline(
    component, pins_function=add_pins, add_outline_function=add_outline
):
    add_outline_function(component)
    pins_function(component)


def add_pins_triangle(component, add_port_marker_function=add_pin_triangle, **kwargs):
    return add_pins(
        component=component, add_port_marker_function=add_port_marker_function, **kwargs
    )


def test_add_pins():
    component = pp.c.mzi2x2(with_elec_connections=True)
    add_pins(component)

    port_layer = port_type2layer["optical"]
    port_markers = read_port_markers(component, [port_layer])
    assert len(port_markers.polygons) == 4

    port_layer = port_type2layer["dc"]
    port_markers = read_port_markers(component, [port_layer])
    assert len(port_markers.polygons) == 3

    # for port_layer, port_type in port_layer2type.items():
    #     port_markers = read_port_markers(component, [port_layer])
    #     print(len(port_markers.polygons))


if __name__ == "__main__":
    test_add_pins()
    # from pp.components import mmi1x2
    # from pp.components import bend_circular
    # from pp.add_grating_couplers import add_grating_couplers

    # c = mmi1x2(width_mmi=5)
    # c = bend_circular()
    # cc = add_grating_couplers(c, layer_label=pp.LAYER.LABEL)

    # c = pp.c.waveguide()
    # c = pp.c.crossing(pins=True)
    # add_pins(c)
    # pp.show(c)



import phidl.geometry as pg
from pp.import_phidl_component import import_phidl_component
from omegaconf.listconfig import ListConfig
from pp.component import Component
from typing import List


def boolean(
    A: Component,
    B: Component,
    operation: str,
    precision: float = 1e-4,
    num_divisions: List[int] = [1, 1],
    max_points: int = 4000,
    layer: ListConfig = 0,
) -> Component:
    """
    Performs boolean operations between 2 Device/DeviceReference objects,
    or lists of Devices/DeviceReferences.

    ``operation`` should be one of {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}.
    Note that 'A+B' is equivalent to 'or', 'A-B' is equivalent to 'not', and
    'B-A' is equivalent to 'not' with the operands switched
    """
    c = pg.boolean(
        A=A,
        B=B,
        operation=operation,
        precision=precision,
        num_divisions=num_divisions,
        max_points=max_points,
        layer=layer,
    )
    return import_phidl_component(component=c)


def _demo():
    import pp

    e1 = pp.c.ellipse()
    e2 = pp.c.ellipse(radii=(10, 6)).movex(2)
    e3 = pp.c.ellipse(radii=(10, 4)).movex(5)
    pp.qp([e1, e2, e3])
    c = boolean(A=[e1, e3], B=e2, operation="A-B")
    pp.show(c)


if __name__ == "__main__":

    _demo()



import pp
from pp.port import deco_rename_ports
from pp.components import waveguide
from pp.components.bend_circular import bend_circular
from pp.components import grating_coupler_elliptical_te
from pp.routing.connect import connect_strip_way_points_no_taper


@pp.autoname
@deco_rename_ports
def _grating_coupler_tree(n_waveguides=4, waveguide_spacing=4, waveguide=waveguide):
    """ array of waveguides connected with grating couplers
    useful to align the 4 corners of the chip
    """

    c = pp.Component()
    w = pp.call_if_func(waveguide)

    for i in range(n_waveguides):
        wref = c.add_ref(w)
        wref.y += i * (waveguide_spacing + w.width)
        c.ports["E" + str(i)] = wref.ports["E0"]
        c.ports["W" + str(i)] = wref.ports["W0"]
    return c


@pp.autoname
def grating_coupler_tree(
    n_waveguides=4,
    waveguide_spacing=4,
    waveguide=waveguide,
    grating_coupler_function=grating_coupler_elliptical_te,
    with_loop_back=False,
    route_filter=connect_strip_way_points_no_taper,
    bend_factory=bend_circular,
    bend_radius=10.0,
    layer_label=pp.LAYER.LABEL,
    **kwargs
):
    """ array of waveguides connected with grating couplers
    useful to align the 4 corners of the chip

    .. plot::
      :include-source:

      import pp

      c = pp.c.grating_coupler_tree()
      pp.plotgds(c)

    """
    c = _grating_coupler_tree(
        n_waveguides=n_waveguides,
        waveguide_spacing=waveguide_spacing,
        waveguide=waveguide,
    )

    cc = pp.routing.add_fiber_array(
        c,
        with_align_ports=with_loop_back,
        optical_routing_type=0,
        grating_coupler=grating_coupler_function,
        bend_radius=bend_radius,
        fanout_length=85.0,
        route_filter=route_filter,
        component_name=c.name,
        straight_factory=waveguide,
        bend_factory=bend_factory,
        layer_label=layer_label,
        **kwargs
    )
    return cc


if __name__ == "__main__":
    c = grating_coupler_tree()
    print(c.ports)
    pp.show(c)



"""
Ports define where each port has:

- name
- midpoint: (x, y)
- width:
- orientation: (deg) 0, 90, 180, 270. where 0 faces east, 90 (north), 180 (west), 270 (south)
- Type:
    - optical
    - electrical (DC)
    - rf (high frequency)
    - detector (Superconducting)
"""


import pp


@pp.autoname
def test_component_with_port(length=5, wg_width=0.5):
    """
    component with one port on the west side
    """

    y = wg_width
    x = length

    c = pp.Component()
    c.add_polygon([(0, 0), (x, 0), (x, y), (0, y)], layer=1)
    c.add_port(
        name="W0",
        midpoint=(0, y / 2),
        width=y,
        orientation=180,
        layer=1,
        port_type="optical",
    )
    assert len(c.ports) == 1
    return c


if __name__ == "__main__":
    c = test_component_with_port()
    pp.show(c)



from typing import Callable, Union
from numpy import ndarray
from phidl.device_layout import Port as PortPhidl

from pp.routing.manhattan import route_manhattan
from pp.routing.manhattan import generate_manhattan_waypoints
from pp.routing.manhattan import round_corners
from pp.components.bend_circular import bend_circular
from pp.components import waveguide
from pp.components import taper as taper_factory
from pp.components.electrical import wire, corner
from pp.component import ComponentReference

from pp.config import WG_EXPANDED_WIDTH, TAPER_LENGTH
from pp.layers import LAYER
from pp.port import Port


def get_waypoints_connect_strip(*args, **kwargs) -> ndarray:
    return connect_strip(*args, **kwargs, route_factory=generate_manhattan_waypoints)


def connect_strip(
    input_port: Port,
    output_port: Union[Port, PortPhidl],
    bend_factory: Callable = bend_circular,
    straight_factory: Callable = waveguide,
    taper_factory: Callable = taper_factory,
    start_straight: float = 0.01,
    end_straight: float = 0.01,
    min_straight: float = 0.01,
    bend_radius: float = 10.0,
    route_factory: Callable = route_manhattan,
) -> ComponentReference:

    bend90 = bend_factory(radius=bend_radius, width=input_port.width)

    if taper_factory:
        if callable(taper_factory):
            taper = taper_factory(
                length=TAPER_LENGTH,
                width1=input_port.width,
                width2=WG_EXPANDED_WIDTH,
                layer=input_port.layer,
            )
        else:
            # In this case the taper is a fixed cell
            taper = taper_factory
    else:
        taper = None

    connector = route_factory(
        input_port,
        output_port,
        bend90,
        straight_factory=straight_factory,
        taper=taper,
        start_straight=start_straight,
        end_straight=end_straight,
        min_straight=min_straight,
    )
    return connector


def connect_strip_way_points(
    way_points=[],
    bend_factory=bend_circular,
    straight_factory=waveguide,
    taper_factory=taper_factory,
    bend_radius=10.0,
    wg_width=0.5,
    layer=LAYER.WG,
    **kwargs
):
    """
    Returns a deep-etched route formed by the given way_points with
    bends instead of corners and optionally tapers in straight sections.

    taper_factory: can be either a taper component or a factory
    """
    bend90 = bend_factory(radius=bend_radius, width=wg_width)

    if taper_factory:
        if callable(taper_factory):
            taper = taper_factory(
                length=TAPER_LENGTH,
                width1=wg_width,
                width2=WG_EXPANDED_WIDTH,
                layer=layer,
            )
        else:
            # In this case the taper is a fixed cell
            taper = taper_factory
    else:
        taper = None

    connector = round_corners(way_points, bend90, straight_factory, taper)
    return connector


def connect_strip_way_points_no_taper(*args, **kwargs):
    return connect_strip_way_points(*args, taper_factory=None, **kwargs)


def connect_elec_waypoints(
    way_points=[],
    bend_factory=corner,
    straight_factory=wire,
    taper_factory=taper_factory,
    wg_width=10.0,
    layer=LAYER.M3,
    **kwargs
):
    """ connect way_points with electrical traces
    """

    bend90 = bend_factory(width=wg_width, layer=layer)

    def _straight_factory(length=10.0, width=wg_width):
        return straight_factory(length=length, width=width, layer=layer)

    connector = round_corners(way_points, bend90, _straight_factory, taper=None)
    return connector


def connect_elec(
    input_port,
    output_port,
    straight_factory=wire,
    bend_factory=corner,
    layer=None,
    **kwargs
):
    width = input_port.width

    # If no layer is specified, the component factories should use the layer
    # given in port.type
    if layer is None:
        layer = input_port.layer

    def _bend_factory(width=width, radius=0):
        return bend_factory(width=width, radius=radius, layer=layer)

    def _straight_factory(length=10.0, width=width):
        return straight_factory(length=length, width=width, layer=layer)

    if "bend_radius" in kwargs:
        bend_radius = kwargs.pop("bend_radius")
    else:
        bend_radius = 10

    return connect_strip(
        input_port,
        output_port,
        bend_radius=bend_radius,
        bend_factory=_bend_factory,
        straight_factory=_straight_factory,
        taper_factory=None,
        **kwargs
    )


if __name__ == "__main__":
    import pp

    w = pp.c.waveguide()

    c = pp.Component()
    connector_ref = connect_strip(w.ports["E0"], w.ports["W0"])
    cc = c.add(connector_ref)
    pp.show(cc)



import uuid
import numpy as np
from pp.components import waveguide
from pp.name import clean_name
from pp.component import Component, ComponentReference

import pp
from pp.geo_utils import angles_deg
from numpy import bool_, float64, ndarray
from typing import Callable, Dict, List, Optional, Tuple
from pp.port import Port

TOLERANCE = 0.0001
DEG2RAD = np.pi / 180
RAD2DEG = 1 / DEG2RAD

O2D = {0: "East", 180: "West", 90: "North", 270: "South"}


def _get_ports_facing(ports: Dict[str, Port], orientation: int = 0) -> List[Port]:
    return [p for p in ports.values() if p.orientation == orientation]


def _get_unique_port_facing(ports: Dict[str, Port], orientation: int = 0) -> List[Port]:
    ports = _get_ports_facing(ports, orientation)

    if len(ports) > 1:
        orientation = orientation % 360
        direction = O2D[orientation]
        raise ValueError(
            "_get_unique_port_facing: \n\
            should have only one port facing {}\n\
            Got {} with names {}".format(
                direction, len(ports), ports
            )
        )

    return ports


def _get_bend_ports(bend: Component) -> List[Port]:
    """
    Any standard bend/corner has two ports: one facing west and one facing north
    Returns these two ports in this order
    """

    ports = bend.ports

    p_w = _get_unique_port_facing(ports, 180)
    p_n = _get_unique_port_facing(ports, 90)

    return p_w + p_n


def _get_straight_ports(straight: Component) -> List[Port]:
    """
    Any standard straight wire/waveguide has two ports:
    one facing west and one facing east
    Returns these two ports in this order
    """
    ports = straight.ports

    p_w = _get_unique_port_facing(ports, 180)
    p_e = _get_unique_port_facing(ports, 0)

    return p_w + p_e


def gen_sref(
    structure: Component,
    rotation_angle: int,
    x_reflection: bool,
    port_name: str,
    position: ndarray,
) -> ComponentReference:
    """
    place sref of `port_name` of `structure` at `position`
    # Keep this convention, otherwise phidl port transform won't work
    # 1 ) Mirror
    # 2 ) Rotate
    # 3 ) Move
    """
    position = np.array(position)

    if port_name is None:
        port_position = np.array([0, 0])
    else:
        port_position = structure.ports[port_name].midpoint

    ref = pp.ComponentReference(component=structure, origin=(0, 0))

    if x_reflection:  # Vertical mirror: Reflection across x-axis
        y0 = port_position[1]
        ref.reflect(p1=(0, y0), p2=(1, y0))

    ref.rotate(rotation_angle, center=port_position)

    ref.move(port_position, position)

    return ref


def _is_vertical(p0: ndarray, p1: ndarray) -> bool_:
    return np.abs(p0[0] - p1[0]) < TOLERANCE


def _is_horizontal(p0: ndarray, p1: ndarray) -> bool_:
    return np.abs(p0[1] - p1[1]) < TOLERANCE


def get_straight_distance(p0: ndarray, p1: ndarray) -> float64:
    if _is_vertical(p0, p1):
        return np.abs(p0[1] - p1[1])
    if _is_horizontal(p0, p1):
        return np.abs(p0[0] - p1[0])

    raise ValueError("Waveguide {} {} is not manhattan".format(p0, p1))


def transform(
    points: ndarray,
    translation: ndarray = (0, 0),
    angle_deg: float64 = 0,
    x_reflection: bool = False,
) -> ndarray:
    """
    Args:
        points (np.array of shape (N,2) ): points to be transformed
        translation (2d like array): translation vector
        angle_deg (float): rotation angle
        x_reflection (bool): if True, mirror the shape across the x axis  (y -> -y)
    """
    # Copy
    pts = points[:, :]

    # Translate
    pts = pts + np.array(translation)

    # Rotate
    c = np.cos(DEG2RAD * angle_deg)
    s = np.sin(DEG2RAD * angle_deg)
    rotation_matrix = np.array([[c, s], [-s, c]])
    pts = np.dot(pts, rotation_matrix)

    # Mirror
    if x_reflection:
        pts[:, 1] = -pts[:, 1]

    return pts


def reverse_transform(
    points: ndarray,
    translation: ndarray = (0, 0),
    angle_deg: float64 = 0,
    x_reflection: bool = False,
) -> ndarray:
    """
    Args:
        points (np.array of shape (N,2) ): points to be transformed
        translation (2d like array): translation vector
        angle_deg (float): rotation angle
        x_reflection (bool): if True, mirror the shape across the x axis  (y -> -y)
    """
    angle_deg = -angle_deg

    # Copy
    pts = points[:, :]

    # Mirror
    if x_reflection:
        pts[:, 1] = -pts[:, 1]

    # Rotate
    c = np.cos(DEG2RAD * angle_deg)
    s = np.sin(DEG2RAD * angle_deg)
    rotation_matrix = np.array([[c, s], [-s, c]])
    pts = np.dot(pts, rotation_matrix)

    # Translate
    pts = pts - np.array(translation)

    return pts


def _generate_route_manhattan_points(
    input_port: Port,
    output_port: Port,
    bs1: float64,
    bs2: float64,
    start_straight: float = 0.01,
    end_straight: float = 0.01,
    min_straight: float = 0.01,
) -> ndarray:
    """
    Args:
        input_port:
        output_port:
        bs1, bs2: float, float : the bend size

    """

    threshold = TOLERANCE

    # transform I/O to the case where output is at (0, 0) pointing east (180)
    p_input = input_port.midpoint
    p_output = output_port.midpoint

    pts_io = np.stack([p_input, p_output], axis=0)

    angle = output_port.orientation

    a0 = -angle + 180
    transform_params = (-p_output, a0, False)

    _pts_io = transform(pts_io, *transform_params)
    p = _pts_io[0, :]
    _p_output = _pts_io[1, :]

    a = int(input_port.orientation + a0) % 360
    s = start_straight
    count = 0
    points = [p]

    while 1:
        count += 1
        if count > 40:
            raise AttributeError(
                "Too many iterations for in {} -> out {}".format(
                    input_port, output_port
                )
            )
        # not ready for final bend: go over possibilities
        sigp = np.sign(p[1])
        if not sigp:
            sigp = 1
        if a % 360 == 0:
            # same directions
            if abs(p[1]) < threshold and p[0] <= threshold:
                # Reach the output!
                points += [_p_output]
                break
            elif (
                p[0] + (bs1 + bs2 + end_straight + s) < threshold
                and abs(p[1]) - (bs1 + bs2 + min_straight) > -threshold
            ):
                # sufficient space for S-bend
                p = (-end_straight - bs2, p[1])
                a = -sigp * 90
            elif (
                p[0] + (2 * bs1 + 2 * bs2 + end_straight + s + min_straight) < threshold
            ):
                # sufficient distance to move aside
                p = (p[0] + s + bs1, p[1])
                a = -sigp * 90
            elif abs(p[1]) - (2 * bs1 + 2 * bs2 + 2 * min_straight) > -threshold:
                p = (p[0] + s + bs1, p[1])
                a = -sigp * 90
            else:
                p = (p[0] + s + bs1, p[1])
                a = sigp * 90

        elif a == 180:
            # opposite directions
            if abs(p[1]) - (bs1 + bs2 + min_straight) > -threshold:
                # far enough: U-turn
                p = (min(p[0] - s, -end_straight) - bs2, p[1])
                a = -sigp * 90
            else:
                # more complex turn
                p = (
                    min(p[0] - s - bs1, -end_straight - min_straight - 2 * bs1 - bs2),
                    p[1],
                )
                a = -sigp * 90
        elif a % 180 == 90:
            siga = -np.sign((a % 360) - 180)
            if not siga:
                siga = 1

            if ((-p[1] * siga) - (s + bs2) > -threshold) and (
                -p[0] - (end_straight + bs2)
            ) > -threshold:
                # simple case: one right angle to the end
                p = (p[0], 0)
                a = 0
            elif (p[1] * siga) <= threshold and p[0] + (
                end_straight + bs1
            ) > -threshold:
                # go to the west, and then turn upward
                # this will sometimes result in too sharp bends, but there is no avoiding this!

                _y = min(
                    max(min(min_straight, 0.5 * abs(p[1])), abs(p[1]) - s - bs1),
                    bs1 + bs2 + min_straight,
                )

                p = (p[0], sigp * _y)
                if count == 1:  # take care of the start_straight case
                    p = (p[0], -sigp * max(start_straight, _y))

                a = 180
            elif -p[0] - (end_straight + 2 * bs1 + bs2 + min_straight) > -threshold:
                # go sufficiently up, and then east
                p = (p[0], siga * max(p[1] * siga + s + bs1, bs1 + bs2 + min_straight))
                a = 0

            elif -p[0] - (end_straight + bs2) > -threshold:
                # make vertical S-bend to get sufficient room for movement
                points += [(p[0], p[1] + siga * (bs2 + s))]
                p = (
                    min(
                        p[0] - bs1 + bs2 + min_straight,
                        -2 * bs1 - bs2 - end_straight - min_straight,
                    ),
                    p[1] + siga * (bs2 + s),
                )
                # `a` remains the same
            else:
                # no viable solution for this case. May result in crossed waveguides
                p = (p[0], p[1] + sigp * (s + bs1))
                a = 180
        points += [p]
        s = min_straight + bs1

    points = np.stack([np.array(_p) for _p in points], axis=0)
    points = reverse_transform(points, *transform_params)
    return points


def _get_bend_reference_parameters(
    p0: ndarray, p1: ndarray, p2: ndarray, bend_cell: Component
) -> Tuple[ndarray, int, bool]:
    """
    8 possible configurations
    First mirror , Then rotate

       p2        p2
        |         |
        | C     A |
       p1-<-p0->-p1      dp1 horizontal
        |         |      dp2 vertical
        | D     B |
       p2        p2


       p2-<-p1->-p2
            |
          F |  E         dp1 vertical
            p0           dp2 horizontal
          H |  G
            |
       p2-<-p1->-p2
    """

    # is_horizontal(dp1), s1, s2 : transform (rotation, vertical mirror)
    transforms_map = {
        (True, 1, 1): (0, False),  # A No transform
        (True, 1, -1): (0, True),  # B Vertical mirror
        (True, -1, 1): (180, True),  # C Horizontal mirror
        (True, -1, -1): (180, False),  # D R180
        (False, 1, 1): (90, True),  # E R90 + vertical mirror
        (False, 1, -1): (90, False),  # F R270
        (False, -1, 1): (270, False),  # G R270
        (False, -1, -1): (270, True),  # H R270 + vertical mirror
    }

    b1, b2 = [p.midpoint for p in _get_bend_ports(bend_cell)]

    bsx = b2[0] - b1[0]
    bsy = b2[1] - b1[1]

    dp1 = p1 - p0
    dp2 = p2 - p1
    is_h_dp1 = np.abs(dp1[1]) < TOLERANCE

    if is_h_dp1:
        xd1 = dp1[0]
        yd2 = dp2[1]
        s1 = np.sign(xd1)
        s2 = np.sign(yd2)

        bend_origin = p1 - np.array([s1 * bsx, 0])

    else:
        yd1 = dp1[1]
        xd2 = dp2[0]
        s1 = int(np.sign(yd1))
        s2 = int(np.sign(xd2))

        bend_origin = p1 - np.array([0, s1 * bsy])

    t = transforms_map[(is_h_dp1, s1, s2)]

    return bend_origin, t[0], t[1]


def make_ref(component_factory: Callable) -> Callable:
    def _make_ref(*args, **kwargs):
        return component_factory(*args, **kwargs).ref()

    return _make_ref


def remove_flat_angles(points: ndarray) -> ndarray:
    a = angles_deg(np.vstack(points))
    da = a - np.roll(a, 1)
    da = np.mod(np.round(da, 3), 180)

    # To make sure we do not remove points at the edges
    da[0] = 1
    da[-1] = 1

    to_rm = list(np.where(np.abs(da[:-1]) < 1e-9)[0])
    if type(points) == list:
        while to_rm:
            i = to_rm.pop()
            points.pop(i)

    else:
        points = points[da != 0]

    return points


@make_ref
def round_corners(
    points,
    bend90,
    straight_factory,
    taper=None,
    straight_factory_fall_back_no_taper=None,
    mirror_straight=False,
    straight_ports=None,
):
    """
    returns a rounded waveguide route from a list of manhattan points
    Args:
        points: manhattan route defined by waypoints
        bend90: the bend to use for 90Deg turns
        straight_factory: the straight factory to use to generate straight portions
        taper: taper for straight portions. If None, no tapering
        straight_factory_fall_back_no_taper: factory to use for straights in case there is no space to put a pair of tapers
        straight_ports: port names for straights. If not specified, will use some heuristic to find them
    """
    ## If there is a taper, make sure its length is known
    if taper:
        if "length" not in taper.info:
            _taper_ports = list(taper.ports.values())
            taper.info["length"] = _taper_ports[-1].x - _taper_ports[0].x

    if straight_factory_fall_back_no_taper is None:
        straight_factory_fall_back_no_taper = straight_factory

    ## Remove any flat angle, otherwise the algorithm won't work
    points = remove_flat_angles(points)

    cell_tmp_name = "connector_{}".format(uuid.uuid4())

    cell = pp.Component(name=cell_tmp_name)

    points = np.array(points)

    straight_sections = []  # (p0, angle, length)
    p0_straight = points[0]
    p1 = points[1]

    total_length = 0  # Keep track of the total path length

    if "length" in bend90.info:
        bend_length = bend90.info["length"]
    else:
        bend_length = 0

    dp = p1 - p0_straight
    a0 = None
    if _is_vertical(p0_straight, p1):
        if dp[1] > 0:
            a0 = 90
        elif dp[1] < 0:
            a0 = 270
    elif _is_horizontal(p0_straight, p1):
        if dp[0] > 0:
            a0 = 0
        elif dp[0] < 0:
            a0 = 180

    assert a0 is not None, "Points should be manhattan, got {} {}".format(
        p0_straight, p1
    )

    pname_west, pname_north = [p.name for p in _get_bend_ports(bend90)]

    n_o_bends = points.shape[0] - 2
    total_length += n_o_bends * bend_length
    # Add bend sections and record straight-section information
    for i in range(1, points.shape[0] - 1):
        bend_origin, rotation, x_reflection = _get_bend_reference_parameters(
            points[i - 1], points[i], points[i + 1], bend90
        )

        bend_ref = gen_sref(bend90, rotation, x_reflection, pname_west, bend_origin)
        cell.add(bend_ref)

        straight_sections += [
            (p0_straight, a0, get_straight_distance(p0_straight, bend_origin))
        ]

        p0_straight = bend_ref.ports[pname_north].midpoint
        a0 = bend_ref.ports[pname_north].orientation

    straight_sections += [
        (p0_straight, a0, get_straight_distance(p0_straight, points[-1]))
    ]

    wg_refs = []
    for straight_origin, angle, length in straight_sections:
        with_taper = False
        wg_width = list(bend90.ports.values())[0].width

        total_length += length

        if taper is not None and length > 2 * taper.info["length"] + 1.0:
            length = length - 2 * taper.info["length"]
            with_taper = True

        if with_taper:

            # First taper:
            # Taper starts where straight would have started
            taper_origin = straight_origin

            pname_west, pname_east = [p.name for p in _get_straight_ports(taper)]
            taper_ref = taper.ref(
                position=taper_origin, port_id=pname_west, rotation=angle
            )

            wg_width = taper.ports[pname_east].width

            cell.add(taper_ref)
            wg_refs += [taper_ref]

            # Update start straight position
            straight_origin = taper_ref.ports[pname_east].midpoint

        # Straight waveguide
        if with_taper or taper is None:
            wg = straight_factory(length=length, width=wg_width)
        else:
            wg = straight_factory_fall_back_no_taper(length=length, width=wg_width)

        if straight_ports is None:
            straight_ports = [p.name for p in _get_straight_ports(wg)]
        pname_west, pname_east = straight_ports

        wg.move(wg.ports[pname_west], (0, 0))
        wg_ref = pp.ComponentReference(wg)
        if mirror_straight:
            wg_ref.reflect_v(list(wg_ref.ports.values())[0].name)

        wg_ref.rotate(angle)
        wg_ref.move(straight_origin)
        cell.add(wg_ref)
        wg_refs += [wg_ref]

        port_index_out = 1
        if with_taper:
            # Second taper:
            # Origin at end of straight waveguide, starting from east side of taper

            taper_origin = wg_ref.ports[pname_east]
            pname_west, pname_east = [p.name for p in _get_straight_ports(taper)]
            taper_ref = taper.ref(
                position=taper_origin, port_id=pname_east, rotation=angle + 180
            )

            cell.add(taper_ref)
            wg_refs += [taper_ref]
            port_index_out = 0

    cell.add_port(name="input", port=list(wg_refs[0].ports.values())[0])
    cell.add_port(name="output", port=list(wg_refs[-1].ports.values())[port_index_out])

    """
    # Update name with uuid - too expensive to compute geometrical hash every time
    # Prefix with zz to make connectors appear at end of cell lists

    # The geometrical hash lacks caching right now and ends up taking a
    # lot of time to compute on every single connector
    """

    cell.name = f"zz_conn_{clean_name(str(uuid.uuid4()))[:16]}"
    cell.info["length"] = total_length
    cell.length = total_length
    return cell


def generate_manhattan_waypoints(
    input_port: Port,
    output_port: Port,
    bend90: Optional[Component] = None,
    bend_radius: None = None,
    start_straight: float = 0.01,
    end_straight: float = 0.01,
    min_straight: float = 0.01,
    **kwargs,
) -> ndarray:
    """

    """

    if bend90 is None and bend_radius is None:
        raise ValueError(
            "Either bend90 or bend_radius must be set. \
        Got {} {}".format(
                bend90, bend_radius
            )
        )

    if bend90 is not None and bend_radius is not None:
        raise ValueError(
            "Either bend90 or bend_radius must be set. \
        Got {} {}".format(
                bend90, bend_radius
            )
        )

    if bend90:
        pname_west, pname_north = [p.name for p in _get_bend_ports(bend90)]
        p1 = bend90.ports[pname_west].midpoint
        p2 = bend90.ports[pname_north].midpoint

        bsx = p2[0] - p1[0]
        bsy = p2[1] - p1[1]

    elif bend_radius:
        bsx = bend_radius
        bsy = bend_radius

    points = _generate_route_manhattan_points(
        input_port, output_port, bsx, bsy, start_straight, end_straight, min_straight
    )
    return points


def route_manhattan(
    input_port: Port,
    output_port: Port,
    bend90: Component,
    straight_factory: Callable,
    taper: None = None,
    start_straight: float = 0.01,
    end_straight: float = 0.01,
    min_straight: float = 0.01,
) -> ComponentReference:
    bend90 = pp.call_if_func(bend90)

    points = generate_manhattan_waypoints(
        input_port,
        output_port,
        bend90=bend90,
        start_straight=start_straight,
        end_straight=end_straight,
        min_straight=min_straight,
    )
    return round_corners(points, bend90, straight_factory, taper)


def test_manhattan():
    from pp.components.bend_circular import bend_circular

    top_cell = pp.Component()

    inputs = [
        Port("in1", (10, 5), 0.5, 90),
        Port("in2", (-10, 20), 0.5, 0),
        Port("in3", (10, 30), 0.5, 0),
        Port("in4", (-10, -5), 0.5, 90),
    ]

    outputs = [
        Port("in1", (90, -60), 0.5, 180),
        Port("in2", (-100, 20), 0.5, 0),
        Port("in3", (100, -25), 0.5, 0),
        Port("in4", (-150, -65), 0.5, 270),
    ]

    for input_port, output_port in zip(inputs, outputs):

        # input_port = Port("input_port", (10,5), 0.5, 90)
        # output_port = Port("output_port", (90,-60), 0.5, 180)

        bend = bend_circular(radius=5.0)
        cell = route_manhattan(
            input_port,
            output_port,
            bend,
            waveguide,
            start_straight=5.0,
            end_straight=5.0,
        )

        top_cell.add(cell)
    return top_cell


if __name__ == "__main__":
    top_cell = test_manhattan()
    pp.show(top_cell)



""" gdsfactory loads a configuration from 3 files, high priority overwrites low priority:

1. A config.yml found in the current working directory (highest priority)
2. ~/.gdsfactory/config.yml specific for the machine
3. the default_config in pp/config.py (lowest priority)

`CONFIG` has all the paths that we do not care
`conf` has all the useful info
"""

__version__ = "1.4.2"
from typing import Any
import os
import io
import json
import subprocess
import pathlib
from pprint import pprint
import logging

import numpy as np
from omegaconf import OmegaConf
from git import Repo


connections = {}  # global variable to store connections in a dict
home = pathlib.Path.home()
cwd = pathlib.Path.cwd()
module_path = pathlib.Path(__file__).parent.absolute()
repo_path = module_path.parent
home_path = pathlib.Path.home() / ".gdsfactory"
home_path.mkdir(exist_ok=True)

cwd_config = cwd / "config.yml"
module_config = module_path / "config.yml"
home_config = home_path / "config.yml"


config_base = OmegaConf.load(
    io.StringIO(
        """
tech:
    name: generic
    cache_url:
    with_settings_label: False
    add_pins: True
    wg_expanded_width: 2.5
    taper_length: 35.0
    grid_unit: 1e-6
    grid_resolution: 1e-9
    bend_radius: 10.0
"""
    )
)


try:
    config_cwd = OmegaConf.load(cwd_config)
except Exception:
    config_cwd = OmegaConf.create()
try:
    config_home = OmegaConf.load(home_config)
except Exception:
    config_home = OmegaConf.create()

conf = OmegaConf.merge(config_base, config_home, config_cwd)
conf.version = __version__

try:
    conf["git_hash"] = Repo(repo_path).head.object.hexsha
except Exception:
    conf["git_hash"] = None


CONFIG = dict(
    config_path=cwd_config.absolute(),
    repo_path=repo_path,
    module_path=module_path,
    gdsdir=module_path / "gds",
    font_path=module_path / "gds" / "alphabet.gds",
    masks_path=repo_path / "mask",
    version=__version__,
    home=home,
    cwd=cwd,
)

mask_name = "notDefined"

if conf.get("mask"):
    mask_name = conf["mask"]["name"]
    mask_config_directory = cwd
    build_directory = mask_config_directory / "build"
    CONFIG["devices_directory"] = mask_config_directory / "devices"
    CONFIG["mask_gds"] = mask_config_directory / "build" / "mask" / (mask_name + ".gds")
else:
    build_directory = home_path / "build"
    mask_config_directory = home_path / "build"

CONFIG["custom_components"] = conf.custom_components
CONFIG["gdslib"] = conf.gdslib or repo_path / "gdslib"
CONFIG["sp"] = CONFIG["gdslib"] / "sp"
CONFIG["gds"] = CONFIG["gdslib"] / "gds"
CONFIG["gdslib_test"] = home_path / "gdslib_test"

CONFIG["build_directory"] = build_directory
CONFIG["gds_directory"] = build_directory / "devices"
CONFIG["cache_doe_directory"] = build_directory / "cache_doe"
CONFIG["doe_directory"] = build_directory / "doe"
CONFIG["mask_directory"] = build_directory / "mask"
CONFIG["mask_gds"] = build_directory / "mask" / (mask_name + ".gds")
CONFIG["mask_config_directory"] = mask_config_directory
CONFIG["gdspath"] = build_directory / "gds.gds"
CONFIG["samples_path"] = module_path / "samples"
CONFIG["netlists"] = module_path / "samples" / "netlists"
CONFIG["components_path"] = module_path / "components"

if "gds_resources" in CONFIG:
    CONFIG["gds_resources"] = CONFIG["masks_path"] / CONFIG["gds_resources"]

build_directory.mkdir(exist_ok=True)
CONFIG["gds_directory"].mkdir(exist_ok=True)
CONFIG["doe_directory"].mkdir(exist_ok=True)
CONFIG["mask_directory"].mkdir(exist_ok=True)
CONFIG["gdslib_test"].mkdir(exist_ok=True)


logging.basicConfig(
    filename=CONFIG["build_directory"] / "log.log",
    filemode="w",
    format="%(name)s - %(levelname)s - %(message)s",
)
logging.warning("This will get logged to a file")


def print_config(key=None):
    if key:
        if CONFIG.get(key):
            print(CONFIG[key])
        else:
            print(f"`{key}` key not found in {cwd_config}")
    else:
        pprint(CONFIG)


def complex_encoder(z):
    if isinstance(z, pathlib.Path):
        return str(z)
    else:
        type_name = type(z)
        raise TypeError(f"Object {z} of type {type_name} is not serializable")


def write_config(config, json_out_path):
    with open(json_out_path, "w") as f:
        json.dump(config, f, indent=2, sort_keys=True, default=complex_encoder)


def call_if_func(f: Any, **kwargs) -> Any:
    return f(**kwargs) if callable(f) else f


def get_git_hash():
    """ Get the current git hash """
    try:
        with open(os.devnull, "w") as shutup:
            return (
                subprocess.check_output(["git", "rev-parse", "HEAD"], stderr=shutup)
                .decode("utf-8")
                .strip("\n")
            )
    except subprocess.CalledProcessError:
        return "not_a_git_repo"


GRID_RESOLUTION = conf.tech.grid_resolution
GRID_PER_UNIT = conf.tech.grid_unit / GRID_RESOLUTION
GRID_ROUNDING_RESOLUTION = int(np.log10(GRID_PER_UNIT))
BEND_RADIUS = conf.tech.bend_radius
TAPER_LENGTH = conf.tech.taper_length
WG_EXPANDED_WIDTH = conf.tech.wg_expanded_width

materials = {
    "si": "Si (Silicon) - Palik",
    "sio2": "SiO2 (Glass) - Palik",
    "sin": "Si3N4 (Silicon Nitride) - Phillip",
}


if __name__ == "__main__":
    # print(conf)
    # print_config("gdslib")
    # print_config()
    # print(CONFIG["git_hash"])
    print(CONFIG["sp"])
    # print(CONFIG)



import pp


def component_with_label():
    c = pp.Component("ellipse_with_label")
    c << pp.c.ellipse()
    c.add_label(text="demo", position=(0.0, 0.0), layer=pp.LAYER.TEXT)
    return c


def test_move():
    """ needs fixing """
    c = component_with_label()
    c.x = 10
    # c.x = 10
    c.movex(10)


if __name__ == "__main__":
    # test_move()
    c = component_with_label()
    c.x = 10.0
    pp.show(c)



""" A Design Of Experiment (DOE) changes one or several component parameters to create a model
"""

import itertools as it
from omegaconf import OmegaConf
from pp.config import CONFIG


sample = """
does:
    cutback_taper_te_400:
      component: cutback_taper_te
      settings:
        wg_width: [0.4]
        length: [1.5]
        n_devices_target: [30, 60, 120]

"""


def load_does(filepath):
    """ returns a dictionary with the information loaded from does.yml

    Args:
        filepath: yaml file describes does

    Returns:
        a dictionnary of DOEs with:
        {
            doe_name1: [(component_factory_name, parameters), ...]
            doe_name2: [(component_factory_name, parameters), ...]
            ...
        }

    .. code:: python

        mmi1x2_gap:
            component: mmi1x2
            gap: [0.5, 0.6]
            length: 10

        mmi1x2:
            component: mmi1x2
            length: [11, 12]
            gap: [0.2, 0.3]
            do_permutation: False

    """
    does = {}
    input_does = OmegaConf.load(filepath)
    input_does = OmegaConf.to_container(input_does)

    for doe_name, doe in input_does.items():
        if doe_name == "mask":
            continue

        if "do_permutation" in doe:
            do_permutation = doe.pop("do_permutation")
        else:
            do_permutation = True
        assert doe.get("settings"), "need to define settings for doe {}".format(
            doe_name
        )

        doe_settings = doe.pop("settings", "")
        if doe_settings:
            doe["settings"] = get_settings_list(do_permutation, **doe_settings)
        else:
            raise ValueError(
                "DOE {doe_name} needs to be a dictionary\n\t got: {doe}\n\t sample: {sample"
            )

        does[doe_name] = doe

    return does


def get_settings_list(do_permutations=True, **kwargs):
    """
    Returns a list of settings

    Args:
        do_permutations: if False, will only zip the values passed for each parameter
        and will not use any combination with default arguments
        **kwargs: Keyword arguments with a list or tuple of desired values to sweep

    Usage:

        import pp

        pp.doe.get_settings_list(length=[30, 40])  # adds different lengths
        pp.doe.get_settings_list(length=[30, 40], width=[4, 8])  # if do_permutations=False, zips arguments (L30W4, L40W8)
        pp.doe.get_settings_list(length=[30, 40], width=[4, 8])  # if do_permutations=True, does all combinations (L30W4, L30W8, L40W4, L40W8)


    add variations of self.baseclass in self.components
    get arguments from default_args and then update them from kwargs
    updates default_args with kwargs
    self.settings lists all the variations
    """

    # Deal with empty parameter case
    if kwargs == {}:
        return {}

    # Accept both values or lists
    for key, value in list(kwargs.items()):
        if not isinstance(value, list):
            kwargs[key] = [value]

    if do_permutations:
        keys, list_values = list(zip(*[x for x in list(kwargs.items())]))
        settings = [dict(list(zip(keys, perms))) for perms in it.product(*list_values)]
    else:
        keys, list_values = list(zip(*[x for x in list(kwargs.items())]))
        settings = [dict(list(zip(keys, values))) for values in zip(*list_values)]

    return settings


def test_load_does():
    filepath = CONFIG["samples_path"] / "mask" / "does.yml"
    does = load_does(filepath)
    assert len(does) == 2
    return does


if __name__ == "__main__":
    test_load_does()
    # from pprint import pprint

    # does_path = CONFIG["samples_path"] / "mask" / "does.yml"
    # pprint(load_does(does_path))



from typing import Callable, Tuple
import phidl.device_layout as pd
from pp.config import call_if_func
from pp.layers import LAYER
from pp.component import Component
from pp.add_labels import get_optical_text
from pp.routing.route_fiber_single import route_fiber_single
from pp.routing.connect import connect_strip_way_points

from pp.components.bend_circular import bend_circular
from pp.components import waveguide
from pp.components.grating_coupler.elliptical_trenches import grating_coupler_te
from pp.container import container
from pp.components.taper import taper
from pp.add_tapers import add_tapers


@container
def add_fiber_single(
    component: Component,
    grating_coupler: Callable = grating_coupler_te,
    layer_label: Tuple[int, int] = LAYER.LABEL,
    optical_io_spacing: int = 50,
    bend_factory: Callable = bend_circular,
    straight_factory: Callable = waveguide,
    taper_factory: Callable = taper,
    taper_length: float = 10.0,
    route_filter: Callable = connect_strip_way_points,
    min_input2output_spacing: int = 127,
    optical_routing_type: int = 2,
    with_align_ports: bool = True,
    component_name=None,
    gc_port_name: str = "W0",
    **kwargs,
) -> Component:
    """returns component with grating ports and labels on each port
    can add align_ports reference structure

    Args:
        component: to connect
        optical_io_spacing: SPACING_GC
        grating_coupler: grating coupler instance, function or list of functions
        bend_factory: bend_circular
        straight_factory: waveguide
        fanout_length: None  # if None, automatic calculation of fanout length
        max_y0_optical: None
        with_align_ports: True, adds loopback structures
        waveguide_separation: 4.0
        bend_radius: BEND_RADIUS
        list_port_labels: None, adds TM labels to port indices in this list
        connected_port_list_ids: None # only for type 0 optical routing
        nb_optical_ports_lines: 1
        force_manhattan: False
        excluded_ports:
        grating_indices: None
        routing_method: connect_strip
        gc_port_name: W0
        optical_routing_type: None: autoselection, 0: no extension
        gc_rotation: -90
        layer_label: LAYER.LABEL
        component_name: name of component
        taper_factory: taper


    .. plot::
      :include-source:

       import pp
       from pp.routing import add_fiber_array

       c = pp.c.crossing()
       cc = add_fiber_array(c)
       pp.plotgds(cc)

    """
    component = component() if callable(component) else component
    grating_coupler = (
        grating_coupler() if callable(grating_coupler) else grating_coupler
    )
    port_width_gc = grating_coupler.ports[gc_port_name].width
    optical_ports = component.get_ports_list(port_type="optical")
    port_width_component = optical_ports[0].width

    if port_width_component != port_width_gc:
        component = add_tapers(
            component,
            taper_factory(
                length=taper_length, width1=port_width_gc, width2=port_width_component
            ),
        )

    elements, grating_couplers, _ = route_fiber_single(
        component,
        optical_io_spacing=optical_io_spacing,
        bend_factory=bend_factory,
        straight_factory=straight_factory,
        route_filter=route_filter,
        grating_coupler=grating_coupler,
        layer_label=layer_label,
        optical_routing_type=optical_routing_type,
        min_input2output_spacing=min_input2output_spacing,
        gc_port_name=gc_port_name,
        **kwargs,
    )

    component_name = component_name or component.name
    name = f"{component_name}_{grating_coupler.polarization}"
    c = Component(name=name)
    cr = c << component
    cr.rotate(90)

    for e in elements:
        c.add(e)
    for gc in grating_couplers:
        c.add(gc)

    for pname, p in component.ports.items():
        if p.port_type != "optical":
            c.add_port(pname, port=p)

    if isinstance(grating_coupler, list):
        grating_couplers = [call_if_func(g) for g in grating_coupler]
        grating_coupler = grating_couplers[0]
    else:
        grating_coupler = call_if_func(grating_coupler)
        grating_couplers = [grating_coupler]

    if with_align_ports:
        gc_port_name = list(grating_coupler.ports.keys())[0]

        length = c.ysize - 2 * grating_coupler.xsize
        wg = c << straight_factory(length=length)
        wg.rotate(90)
        wg.xmax = (
            c.xmin - optical_io_spacing
            if abs(c.xmin) > abs(optical_io_spacing)
            else c.xmin - optical_io_spacing
        )
        wg.ymin = c.ymin + grating_coupler.xsize

        gci = c << grating_coupler
        gco = c << grating_coupler
        gci.connect(gc_port_name, wg.ports["W0"])
        gco.connect(gc_port_name, wg.ports["E0"])

        gds_layer_label, gds_datatype_label = pd._parse_layer(layer_label)

        port = wg.ports["E0"]
        text = get_optical_text(
            port, grating_coupler, 0, component_name=f"loopback_{component.name}"
        )

        label = pd.Label(
            text=text,
            position=port.midpoint,
            anchor="o",
            layer=gds_layer_label,
            texttype=gds_datatype_label,
        )
        c.add(label)

        port = wg.ports["W0"]
        text = get_optical_text(
            port, grating_coupler, 1, component_name=f"loopback_{component.name}"
        )
        label = pd.Label(
            text=text,
            position=port.midpoint,
            anchor="o",
            layer=gds_layer_label,
            texttype=gds_datatype_label,
        )
        c.add(label)

    return c


if __name__ == "__main__":
    import pp

    c = pp.c.crossing()
    c = pp.c.mmi2x2()
    c = pp.c.ring_double(length_y=3)  # FIXME
    c = pp.c.waveguide(width=2, length=500)

    cc = add_fiber_single(c)
    print(cc.get_settings()["component"])
    pp.show(cc)



import pp
from pp.component import Component
from typing import Dict, List, Tuple

DIRECTION_TO_ANGLE = {"W": 180, "E": 0, "N": 90, "S": 270}


@pp.autoname
def rectangle(
    size: Tuple[float, float] = (4.0, 2.0),
    layer: Tuple[int, int] = pp.LAYER.WG,
    centered: bool = False,
    ports_parameters: Dict[str, List[Tuple[float, float]]] = {},
    **port_settings
) -> Component:
    """ rectangle

    Args:
        size: (tuple) Width and height of rectangle.
        layer: (int, array-like[2], or set) Specific layer(s) to put polygon geometry on.
        ports: {direction: [(x_or_y, width), ...]} direction: 'W', 'E', 'N' or 'S'

    .. plot::
      :include-source:

      import pp

      c = pp.c.rectangle(size=(4, 2), layer=0)
      pp.plotgds(c)
    """

    c = pp.Component()
    w, h = size

    if centered:
        points = [
            [-w / 2.0, -h / 2.0],
            [-w / 2.0, h / 2],
            [w / 2, h / 2],
            [w / 2, -h / 2.0],
        ]
    else:
        points = [[w, h], [w, 0], [0, 0], [0, h]]
    c.add_polygon(points, layer=layer)

    i = 0
    for direction, list_port_params in ports_parameters.items():
        angle = DIRECTION_TO_ANGLE[direction]
        for x_or_y, width in list_port_params:
            if direction == "W":
                position = (0, x_or_y)

            elif direction == "E":
                position = (w, x_or_y)

            elif direction == "S":
                position = (x_or_y, 0)

            elif direction == "N":
                position = (x_or_y, h)

            c.add_port(
                name="{}".format(i),
                orientation=angle,
                midpoint=position,
                width=width,
                layer=layer,
                **port_settings
            )
            i += 1

    pp.port.auto_rename_ports(c)
    return c


@pp.autoname
def rectangle_centered(
    w: int = 1, h: int = 1, x: None = None, y: None = None, layer: int = 0
) -> Component:
    """ a rectangle size (x, y) in layer
        bad naming with x and y. Replaced with w and h. Keeping x and y
        for now for backwards compatibility

    .. plot::
      :include-source:

      import pp

      c = pp.c.rectangle_centered(w=1, h=1, layer=0)
      pp.plotgds(c)
    """
    c = pp.Component()
    if x:
        w = x
    if y:
        h = y

    points = [
        [-w / 2.0, -h / 2.0],
        [-w / 2.0, h / 2],
        [w / 2, h / 2],
        [w / 2, -h / 2.0],
    ]
    c.add_polygon(points, layer=layer)
    return c


if __name__ == "__main__":
    # c = rectangle_centered()
    c = rectangle()
    print(c.ports)
    pp.show(c)



"""
This is a sample on how to define custom components.
You can make a repo out of this file, having one custom component per file
"""
import os
import shutil
import pytest
import pp
from pp.config import CONFIG
from pp.components.spiral_inner_io import spiral_inner_io_euler
from pp.add_termination import add_gratings_and_loop_back
from pp.routing.connect import connect_strip_way_points
from pp.add_padding import add_padding_to_grid

from pp.mask.merge_metadata import merge_metadata


def _route_filter(*args, **kwargs):
    return connect_strip_way_points(
        *args, taper_factory=None, start_straight=5.0, end_straight=5.0, **kwargs
    )


def add_te(component, **kwargs):
    c = pp.routing.add_fiber_array(
        component,
        grating_coupler=pp.c.grating_coupler_elliptical_te,
        route_filter=_route_filter,
        **kwargs,
    )
    c.test = "passive_optical_te"
    c = add_padding_to_grid(c)
    return c


def add_tm(component, **kwargs):
    c = pp.routing.add_fiber_array(
        component,
        grating_coupler=pp.c.grating_coupler_elliptical_tm,
        route_filter=_route_filter,
        bend_radius=20,
        **kwargs,
    )
    c = add_padding_to_grid(c)
    return c


@pp.autoname
def coupler_te(gap, length, wg_width=0.5, nominal_wg_width=0.5):
    """ sample of component cutback """
    c = pp.c.coupler(wg_width=wg_width, gap=gap, length=length)
    cc = add_te(c)
    return cc


@pp.autoname
def spiral_te(wg_width=0.5, length=2):
    """ sample of component cutback

    Args:
        wg_width: um
        lenght: mm
    """
    c = spiral_inner_io_euler(wg_width=wg_width, length=length)
    cc = add_gratings_and_loop_back(
        component=c,
        grating_coupler=pp.c.grating_coupler_elliptical_te,
        bend_factory=pp.c.bend_circular,
    )
    return cc


@pp.autoname
def spiral_tm(wg_width=0.5, length=2):
    """ sample of component cutback """
    c = spiral_inner_io_euler(wg_width=wg_width, length=length, dx=10, dy=10, N=5)
    cc = add_gratings_and_loop_back(
        component=c,
        grating_coupler=pp.c.grating_coupler_elliptical_tm,
        bend_factory=pp.c.bend_circular,
    )
    return cc


@pytest.fixture
def cleandir():
    build_folder = CONFIG["samples_path"] / "mask_custom" / "build"
    if build_folder.exists():
        shutil.rmtree(build_folder)


@pytest.fixture
def chdir():
    workspace_folder = CONFIG["samples_path"] / "mask_custom"
    os.chdir(workspace_folder)


@pytest.mark.usefixtures("cleandir")
def test_mask(precision=1e-9):
    workspace_folder = CONFIG["samples_path"] / "mask_pack"
    build_path = workspace_folder / "build"
    mask_path = build_path / "mask"

    mask_path.mkdir(parents=True, exist_ok=True)

    gdspath = mask_path / "sample_mask.gds"
    markdown_path = gdspath.with_suffix(".md")
    json_path = gdspath.with_suffix(".json")
    test_metadata_path = gdspath.with_suffix(".tp.json")

    components = [spiral_te(length=length) for length in [2, 4, 6]]
    components += [coupler_te(length=length, gap=0.2) for length in [10, 20, 30, 40]]
    c = pp.pack(components)
    m = c[0]
    m.name = "sample_mask"
    pp.write_gds(m, gdspath)

    merge_metadata(gdspath=gdspath)

    assert gdspath.exists()
    assert markdown_path.exists()
    assert json_path.exists()
    assert test_metadata_path.exists()
    return gdspath


if __name__ == "__main__":
    c = test_mask()
    pp.klive.show(c)



import pp
from pp.components.coupler import coupler

# from pp.components.loop_mirror import loop_mirror
from pp.container import container
from pp.component import Component


@container
@pp.autoname
def cavity(
    component: Component,
    coupler: Component = coupler,
    length: float = 0.1,
    gap: float = 0.2,
    wg_width: float = 0.5,
) -> Component:
    """ creates a cavity from a coupler and a mirror
    it will connect the W0 port of the mirror to both E1 and W1 ports of the coupler creating a resonant cavity

    Args:
        component: mirror
        coupler: coupler factory
        length: coupler length
        gap: coupler gap
        wg_width: coupler wg_width

    .. code::

      ml (mirror left)              mr (mirror right)
       |                               |
       |W0 - W1__             __E1 - W0|
       |         \           /         |
                  \         /
                ---=========---
             W0    length      E0

    .. plot::
      :include-source:

      import pp

      c = pp.c.cavity(component=pp.c.dbr())
      pp.plotgds(c)
    """
    mirror = pp.call_if_func(component)
    coupler = pp.call_if_func(coupler, length=length, gap=gap, wg_width=wg_width)

    c = pp.Component()
    cr = c << coupler
    ml = c << mirror
    mr = c << mirror

    ml.connect("W0", destination=cr.ports["W1"])
    mr.connect("W0", destination=cr.ports["E1"])
    c.add_port("W0", port=cr.ports["W0"])
    c.add_port("E0", port=cr.ports["E0"])
    return c


if __name__ == "__main__":
    from pp.components.dbr import dbr

    c = cavity(component=dbr(), pins=True)
    pp.show(c)



"""
# Metadata

Together with the GDS files that we send to the foundries we also store some .JSON dictionaries for each cell containing all the settings that we used to build the GDS.

By default the metadata will consists of all the parameters that were passed to the component function.
"""


import pp

c = pp.c.waveguide()

print(c.settings)
print(c.get_json())



""" Picwriter is a photonics library written by Derek Kita
https://picwriter.readthedocs.io/en/latest/component-documentation.html
"""
import numpy as np
import gdspy

import picwriter.toolkit as pt
import picwriter.components as pc
import pp
from pp.component import Component


gdspy.current_library = gdspy.GdsLibrary()


def direction_to_degree(direction: str) -> float:
    """ Converts a 'direction' (as used in picwriter) to an angle in degrees.
    picwriter 'direction's can be either a float (corresponding to an angle in radians)
    or a string, corresponding to a cardinal direction
    """
    if isinstance(direction, float):
        # direction is a float in radians, but rotation should be a float in degrees
        return direction * 180.0 / np.pi
    elif str(direction) == "EAST":
        return 0.0
    elif str(direction) == "NORTH":
        return 90.0
    elif str(direction) == "WEST":
        return 180.0
    elif str(direction) == "SOUTH":
        return 270.0


def picwriter2component(picwriter_object: pt.Component) -> Component:
    """ Converts a Picwriter into a Gdsfactory Component
    """
    po = picwriter_object
    c = pp.Component(name=po.name_prefix)

    # Add the polygons
    po_cell = pt.CURRENT_CELLS[
        po.cell_hash
    ]  # Extract the relevant cells from the picwriter global cell list

    ps = po_cell.get_polygonsets()
    for i in range(len(ps)):
        polygons = ps[i].polygons
        layers = ps[i].layers
        datatypes = ps[i].datatypes

        for j in range(len(polygons)):
            c.add_polygon(polygons[j], layer=(layers[j], datatypes[j]))

    translate_by = po.port
    rotate_by = direction_to_degree(po.direction)

    c.rotate(rotate_by)  # First rotate about (0,0)
    c.move(translate_by)  # Next translate

    """ Add the ports """
    for port in po.portlist.keys():
        port_loc = po.portlist[port]["port"]
        direction = direction_to_degree(po.portlist[port]["direction"])

        c.add_port(
            name=port,
            midpoint=[port_loc[0], port_loc[1]],
            width=po.wgt.wg_width,
            orientation=direction,
        )

    return c


if __name__ == "__main__":

    wgt = pc.WaveguideTemplate(
        bend_radius=50.0,
        wg_width=1.0,
        wg_layer=1,
        wg_datatype=0,
        clad_layer=2,
        clad_datatype=0,
    )

    # gc = pc.GratingCoupler(wgt, port=(10, 20), direction=np.pi * 7 / 8)
    gc = pc.GratingCoupler(wgt, port=(10, 20), direction=0.0)
    gcc = picwriter2component(gc)

    pp.show(gcc)



import numpy as np
import pp
from pp.component import Component


@pp.autoname
def verniers(
    width_min: float = 0.1, width_max: float = 0.5, gap: float = 0.1, size_max: int = 11
) -> Component:
    c = pp.Component()
    y = 0

    widths = np.linspace(width_min, width_max, int(size_max / (width_max + gap)))

    for width in widths:
        w = c << pp.c.waveguide(width=width, length=size_max, layers_cladding=[])
        y += width / 2
        w.y = y
        c.add(pp.c.label(str(int(width * 1e3)), position=(0, y)))
        y += width / 2 + gap

    return c


if __name__ == "__main__":
    c = verniers()
    c.flatten()
    pp.write_gds(c, "verniers.gds")
    pp.show(c)



from typing import Any, List, Tuple
from numpy import floor
from pp.layers import LAYER
from pp.name import autoname
from pp.component import Component


@autoname
def via(width=0.7, height=0.7, period=2.0, clearance=1.0, layer=LAYER.VIA1):
    """ Rectangular via
    """
    c = Component()
    c.info["period"] = period
    c.info["clearance"] = clearance
    c.info["width"] = width
    c.info["height"] = height

    a = width / 2
    b = height / 2

    c.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)

    return c


@autoname
def via1(**kwargs):
    return via(layer=LAYER.VIA1, **kwargs)


@autoname
def via2(**kwargs):
    return via(layer=LAYER.VIA2, **kwargs)


@autoname
def via3(**kwargs):
    return via(layer=LAYER.VIA3, **kwargs)


@autoname
def tlm(
    width: float = 11.0,
    height: float = 11.0,
    layers: List[Tuple[int, int]] = [LAYER.M1, LAYER.M2, LAYER.M3],
    vias: List[Any] = [via2, via3],
) -> Component:
    """
    Rectangular transition thru metal layers

    Args:
        name: component name
        width, height: rectangle parameters
        layers: layers on which to draw rectangles
        vias: vias to use to fill the rectangles

    Returns
        <Component>
    """

    # assert len(layers) - 1 == len(vias), "tlm: There should be N layers for N-1 vias"

    a = width / 2
    b = height / 2
    rect_pts = [(-a, -b), (a, -b), (a, b), (-a, b)]

    c = Component()
    # Add metal rectangles
    for layer in layers:
        c.add_polygon(rect_pts, layer=layer)

    # Add vias
    for via in vias:
        via = via() if callable(via) else via

        w = via.info["width"]
        h = via.info["height"]
        g = via.info["clearance"]
        period = via.info["period"]

        nb_vias_x = (width - w - 2 * g) / period + 1
        nb_vias_y = (height - h - 2 * g) / period + 1

        nb_vias_x = int(floor(nb_vias_x))
        nb_vias_y = int(floor(nb_vias_y))

        cw = (width - (nb_vias_x - 1) * period - w) / 2
        ch = (height - (nb_vias_y - 1) * period - h) / 2

        x0 = -a + cw + w / 2
        y0 = -b + ch + h / 2

        for i in range(nb_vias_x):
            for j in range(nb_vias_y):
                c.add(via.ref(position=(x0 + i * period, y0 + j * period)))

    return c


if __name__ == "__main__":
    import pp

    c = tlm()
    pp.show(c)



import pp
from pp.layers import LAYER
from pp.component import Component
from typing import Tuple


@pp.autoname
def hline(
    length: float = 10.0, width: float = 0.5, layer: Tuple[int, int] = LAYER.WG,
) -> Component:
    """ horizonal line waveguide, with ports on east and west sides

    .. plot::
      :include-source:

      import pp
      c = pp.c.hline()
      pp.plotgds(c)

    """
    c = pp.Component()
    a = width / 2
    if length > 0 and width > 0:
        c.add_polygon([(0, -a), (length, -a), (length, a), (0, a)], layer=layer)

    c.add_port(name="W0", midpoint=[0, 0], width=width, orientation=180, layer=layer)
    c.add_port(name="E0", midpoint=[length, 0], width=width, orientation=0, layer=layer)

    c.width = width
    c.length = length
    return c


if __name__ == "__main__":
    c = hline(width=10)
    print(c)
    pp.show(c)



""" CD SEM structures
"""

from typing import Callable, List
import itertools as it
from omegaconf.listconfig import ListConfig
import numpy as np
import pp
from pp.components.bend_circular import bend_circular
from pp.components.waveguide import waveguide
from pp.layers import LAYER
from pp.port import rename_ports_by_orientation
from pp.components.manhattan_font import manhattan_text
from pp.component import Component

LINE_LENGTH = 420.0


@pp.autoname
def square_middle(
    side: float = 0.5,
    layer: ListConfig = LAYER.WG,
    cladding_offset: int = 3,
    layers_cladding: List[ListConfig] = [],
) -> Component:
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, -a), (a, -a), (a, a), (-a, a)], layer=layer)
    a += cladding_offset
    for layer_cladding in layers_cladding:
        component.add_polygon(
            [(-a, -a), (a, -a), (a, a), (-a, a)], layer=layer_cladding
        )
    return component


@pp.autoname
def double_square(side=0.5, layer=LAYER.WG, layers_cladding=[], cladding_offset=3):
    component = pp.Component()
    a = side / 2
    pts0 = [(-a, -a), (a, -a), (a, a), (-a, a)]
    for b in [-side / 2 - 0.25, side / 2 + 0.25]:
        pts = [(x, y + b) for x, y in pts0]
        component.add_polygon(pts, layer=layer)

    a += cladding_offset
    for layer in layers_cladding:
        component.add_polygon([(-a, -a), (a, -a), (a, a), (-a, a)], layer=layer)

    return component


@pp.autoname
def rectangle(
    x: float,
    y: float,
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
    cladding_offset: int = 3,
) -> Component:
    component = pp.Component()
    a = x / 2
    b = y / 2
    component.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)
    a += cladding_offset
    b += cladding_offset
    for layer in layers_cladding:
        component.add_polygon([(-a, -b), (a, -b), (a, b), (-a, b)], layer=layer)
    return component


def triangle_middle_up(side=0.5, layer=LAYER.WG):
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, -a), (a, -a), (0, a)], layer=layer)
    return component


def triangle_middle_down(side=0.5, layer=LAYER.WG):
    component = pp.Component()
    a = side / 2
    component.add_polygon([(-a, a), (a, a), (0, -a)], layer=layer)
    return component


@pp.autoname
def char_H(
    layer: ListConfig = LAYER.WG, layers_cladding: List[ListConfig] = []
) -> Component:
    return manhattan_text("H", size=0.4, layer=layer, layers_cladding=layers_cladding)


@pp.autoname
def char_L(
    layer: ListConfig = LAYER.WG, layers_cladding: List[ListConfig] = []
) -> Component:
    return manhattan_text("L", size=0.4, layer=layer, layers_cladding=layers_cladding)


CENTER_SHAPES_MAP = {
    "S": square_middle,
    "U": double_square,
    "D": double_square,
    "H": char_H,
    "L": char_L,
}


def _cdsem_generic(
    L=2.0,
    radius=2.0,
    width=0.4,
    center_shapes="SU",
    bend90_factory=bend_circular,
    waveguide_factory=waveguide,
    markers_with_slabs=False,
):
    """bends and straights connected together
    for CDSEM measurement
    """

    component = pp.Component()
    _straight = waveguide_factory(length=L, width=width)
    _bend = bend90_factory(radius=radius, width=width)

    straight1 = _straight.ref(rotation=90, port_id="W0")
    component.add(straight1)

    bend1 = component.add_ref(_bend)
    bend1.connect(port="N0", destination=straight1.ports["E0"])

    straight2 = component.add_ref(_straight)
    straight2.connect(port="W0", destination=bend1.ports["W0"])

    bend2 = component.add_ref(_bend)
    bend2.connect(port="N0", destination=straight2.ports["E0"])

    # Add center shapes.
    # Center the first shape in the list
    # Then stack the others underneath

    center = np.array([radius + L / 2, L / 2])
    center_shape_side = 0.4 if markers_with_slabs else 0.5
    center_shape_spacing = 0.2
    sep = center_shape_side + center_shape_spacing
    for cs_name in center_shapes:
        _shape_func = CENTER_SHAPES_MAP[cs_name]
        _shape = _shape_func(side=center_shape_side)
        _shape_ref = _shape.ref(position=center)
        component.add(_shape_ref)
        component.absorb(_shape_ref)

        # If with slabs, add the square slab on top of each marker
        if markers_with_slabs:
            _shape = square_middle(side=0.5)
            _shape_ref = _shape.ref(position=center)
            component.add(_shape_ref)
            component.absorb(_shape_ref)

        center += (0, -sep)

    return component


def wg_line(
    length: float,
    width: float,
    offset: float = 0.2,
    layer: ListConfig = pp.LAYER.WG,
    layers_cladding: List[ListConfig] = [],
) -> Component:
    c = pp.Component()
    _wg = c.add_ref(
        rectangle(x=length, y=width, layer=layer, layers_cladding=layers_cladding)
    )
    c.absorb(_wg)
    return c


@pp.autoname
def cdsem_straight(
    name=None,
    spacing_h=5.0,
    spacing_v=5.0,
    gaps=[0.224, 0.234, 0.246],
    length=10.0,
    width_center=0.5,
    label="A",
    waveguide_factory=waveguide,
    layer=LAYER.WG,
    layers_cladding=[],
):

    c = pp.Component()
    x = 0
    i = 0

    widths = [width_center * 0.92, width_center, width_center * 1.08]

    marker_label = manhattan_text(
        text=label, size=0.4, layer=layer, layers_cladding=layers_cladding
    )
    _marker_label = c.add_ref(marker_label)
    _marker_label.move((2 * (length + spacing_h) + 2 * spacing_h, 1.5 * spacing_v))
    c.absorb(_marker_label)

    for width, marker_type in zip(widths, ["L", "S", "H"]):
        y = 0
        # iso line
        _r = wg_line(length, width, layer=layer, layers_cladding=layers_cladding)
        _r_ref = c.add_ref(_r)
        _r_ref.move((x, y))
        c.absorb(_r_ref)

        y += width + spacing_v
        # dual lines
        for gap, col_marker_type in zip(gaps, ["S", "L", "H"]):
            wg_line(length, width, layer=layer, layers_cladding=layers_cladding)
            _r1_ref = c.add_ref(_r)
            _r1_ref.move((x, y))

            waveguide_factory(length=length, width=width)
            _r2_ref = c.add_ref(_r)
            _r2_ref.move((x, y + gap + width))
            c.absorb(_r1_ref)
            c.absorb(_r2_ref)

            if i < 2:
                marker = CENTER_SHAPES_MAP[col_marker_type](
                    layer=layer, layers_cladding=layers_cladding
                )
                _marker = c.add_ref(marker)
                _marker.move((x + length / 2 + spacing_h / 2, y + width + gap / 2))
                c.absorb(_marker)

            y += 2 * width + gap + spacing_v
        i += 1

        x += length + spacing_h

    c.move(c.size_info.cc, (0, 0))
    return c


@pp.autoname
def cdsem_straight_column(
    name: None = None,
    spacing_h: float = 5.0,
    spacing_v: float = 5.0,
    gaps: List[float] = [0.224, 0.234, 0.246],
    length: float = LINE_LENGTH,
    width_center: float = 0.5,
    label: str = "A",
    waveguide_factory: Callable = waveguide,
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
) -> Component:

    c = pp.Component()
    x = 0

    widths = [width_center * 0.92, width_center, width_center * 1.08]

    y = 0
    for j, width in enumerate(widths):
        # iso line
        _r = waveguide_factory(length=length, width=width)
        _r_ref = c.add_ref(_r)
        _r_ref.move((x, y))
        c.absorb(_r_ref)

        y += width + spacing_v

        marker_label = manhattan_text(
            text="{}{}".format(label, j + 1),
            size=0.4,
            layer=layer,
            layers_cladding=layers_cladding,
        )
        _marker_label = c.add_ref(marker_label)
        _marker_label.move((length + 6, y))
        c.absorb(_marker_label)

        # dual lines
        for gap, col_marker_type in zip(gaps, ["S", "L", "H"]):
            wg_line(length, width, layer=layer, layers_cladding=layers_cladding)
            _r1_ref = c.add_ref(_r)
            _r1_ref.move((x, y))

            wg_line(length, width, layer=layer, layers_cladding=layers_cladding)
            _r2_ref = c.add_ref(_r)
            _r2_ref.move((x, y + gap + width))
            c.absorb(_r1_ref)
            c.absorb(_r2_ref)

            # if i < 2:
            marker = CENTER_SHAPES_MAP[col_marker_type](
                layer=layer, layers_cladding=layers_cladding
            )
            _marker = c.add_ref(marker)
            _marker.move((length + 3, y + width + gap / 2))
            c.absorb(_marker)

            y += 2 * width + gap + spacing_v
        # i += 1

        y += spacing_v + 2 * width

    c.move(c.size_info.cc, (0, 0))
    return c


@pp.autoname
def cdsem_straight_all(
    waveguide_factory: Callable = waveguide,
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
) -> Component:
    widths = [0.4, 0.45, 0.5, 0.6, 0.8, 1.0]
    labels = ["A", "B", "C", "D", "E", "F"]
    c = pp.Component()
    spacing_v = 10.0
    y = 0
    for width, label in zip(widths, labels):
        _c = cdsem_straight_column(
            width_center=width,
            label=label,
            waveguide_factory=waveguide_factory,
            layer=layer,
            layers_cladding=layers_cladding,
        )
        y -= _c.size_info.south
        cr = c.add_ref(_c)
        cr.movey(y)
        y = cr.size_info.north + spacing_v

    return c


@pp.autoname
def cdsem_straight_density(
    wg_width: float = 0.372,
    trench_width: float = 0.304,
    x: float = LINE_LENGTH,
    y: float = 50.0,
    margin: float = 2.0,
    label: str = "",
    waveguide_factory: Callable = waveguide,
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
) -> Component:
    """horizontal grating etch lines

    TE: 676nm pitch, 304nm gap, 372nm line
    TM: 1110nm pitch, 506nm gap, 604nm line

    Args:
        w: wg_width
        s: trench_width
    """
    c = pp.Component()
    period = wg_width + trench_width
    n_o_lines = int((y - 2 * margin) / period)
    length = x - 2 * margin

    tooth = waveguide_factory(length=length, width=wg_width)

    for i in range(n_o_lines):
        tooth_ref = c.add_ref(tooth)
        tooth_ref.movey((-n_o_lines / 2 + 0.5 + i) * period)
        c.absorb(tooth_ref)

    marker_label = manhattan_text(
        text="{}".format(label), size=1.0, layer=layer, layers_cladding=layers_cladding
    )
    _marker_label = c.add_ref(marker_label)
    _marker_label.move((length + 3, 10.0))
    c.absorb(_marker_label)

    return c


@pp.autoname
def cdsem_strip(waveguide_factory=waveguide, **kwargs):
    return _cdsem_generic(
        **kwargs, bend90_factory=bend_circular, waveguide_factory=waveguide_factory
    )


@pp.autoname
def cdsem_target(
    bend90_factory: Callable = bend_circular,
    width_center: float = 0.5,
    label: str = "",
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
    radii: List[float] = [5.0, 10.0],
) -> Component:
    c = pp.Component()
    a = 1.0
    w = 3 * a / 4

    for pos in [(0, 0), (w, w), (-w, w), (w, -w), (-w, -w)]:
        _c = c.add_ref(square_middle(layer=layer, layers_cladding=layers_cladding))
        _c.move(pos)
        c.absorb(_c)

    w_min = width_center * 0.92
    w0 = width_center
    w_max = width_center * 1.08

    for radius in radii:
        b = a + radius
        _b_tr = bend90_factory(radius=radius, width=w0)
        b_tr = _b_tr.ref(position=(b, a), rotation=90, port_id="W0")

        _b_bl = bend90_factory(radius=radius, width=w0)
        b_bl = _b_bl.ref(position=(-b, -a), rotation=270, port_id="W0")

        _b_br = bend90_factory(radius=radius, width=w_max)
        b_br = _b_br.ref(position=(a, -b), rotation=0, port_id="W0")

        _b_tl = bend90_factory(radius=radius, width=w_min)
        b_tl = _b_tl.ref(position=(-a, b), rotation=180, port_id="W0")

        c.add([b_tr, b_tl, b_bl, b_br])

    if label:
        marker_label = manhattan_text(text="{}".format(label), size=1.0, layer=layer)
        _marker_label = c.add_ref(marker_label)
        _marker_label.movey(-max(radii) - 10.0)
        c.absorb(_marker_label)

    return c


@pp.autoname
def cdsem_uturn(
    width: float = 0.5,
    radius: float = 10.0,
    symbol_bot: str = "S",
    symbol_top: str = "U",
    wg_length: float = LINE_LENGTH,
    waveguide_factory: Callable = pp.c.waveguide,
    bend90_factory: Callable = bend_circular,
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
) -> Component:
    """

    Args:
        width: of the line
        cladding_offset:
        radius: bend radius
        wg_length

    """
    print(bend90_factory)
    c = pp.Component()
    r = radius
    bend90 = bend90_factory(width=width, radius=r)
    if wg_length is None:
        wg_length = 2 * r
    wg = waveguide_factory(width=width, length=wg_length,)

    # bend90.ports()
    rename_ports_by_orientation(bend90)

    # Add the U-turn on waveguide layer
    b1 = c.add_ref(bend90)
    b2 = c.add_ref(bend90)

    b2.connect("N0", b1.ports["W0"])

    wg1 = c.add_ref(wg)
    wg1.connect("W0", b1.ports["N0"])

    wg2 = c.add_ref(wg)
    wg2.connect("W0", b2.ports["W0"])

    # Add symbols

    sym1 = c.add_ref(
        CENTER_SHAPES_MAP[symbol_bot](layer=layer, layers_cladding=layers_cladding)
    )
    sym1.rotate(-90)
    sym1.movey(r)
    sym2 = c.add_ref(
        CENTER_SHAPES_MAP[symbol_top](layer=layer, layers_cladding=layers_cladding)
    )
    sym2.rotate(-90)
    sym2.movey(2 * r)
    c.absorb(sym1)
    c.absorb(sym2)

    c.rotate(angle=90)
    # print(c._bb_valid)
    # print(c.size_info)
    c.move(c.size_info.cc, (0, 0))
    return c


@pp.autoname
def opcm(
    dw: float = 0.02,
    wte: float = 0.372,
    tte: float = 0.304,
    wtm: float = 0.604,
    ttm: float = 0.506,
    waveguide_factory: Callable = waveguide,
    bend90_factory: Callable = bend_circular,
    layer: ListConfig = LAYER.WG,
    layers_cladding: List[ListConfig] = [],
) -> Component:
    """column with all optical PCMs
    Args:
        dw
    """
    c = pp.Component()
    spacing_v = 5.0
    _c1 = cdsem_straight_all(
        waveguide_factory=waveguide_factory,
        layer=layer,
        layers_cladding=layers_cladding,
    )

    all_devices = [_c1]

    all_devices += [
        cdsem_uturn(
            width=w,
            symbol_top=s,
            waveguide_factory=waveguide_factory,
            bend90_factory=bend90_factory,
            layer=layer,
            layers_cladding=layers_cladding,
        )
        for w, s in zip([0.46, 0.5, 0.54], ["L", "S", "H"])
    ]

    density_params = [
        (wte - dw, tte - dw, "AL"),
        (wte, tte, "AM"),
        (wte + dw, tte + dw, "AH"),
    ]
    density_params += [
        (wtm - dw, ttm - dw, "BL"),
        (wtm, ttm, "BM"),
        (wtm + dw, ttm + dw, "BH"),
    ]

    all_devices += [
        cdsem_straight_density(
            wg_width=w,
            trench_width=t,
            label=lbl,
            waveguide_factory=waveguide_factory,
            layer=layer,
            layers_cladding=layers_cladding,
        )
        for w, t, lbl in density_params
    ]

    y = 0
    for d in all_devices:
        c.add_ref(d)
        y = y - d.ymin
        x = -d.xmin
        d.movey(y)
        d.movex(x)
        y = d.ymax + spacing_v

    widths = [0.4, 0.45, 0.5, 0.6, 0.8, 1.0]
    labels = ["A", "B", "C", "D", "E", "F"]
    targets = [
        cdsem_target(
            bend90_factory=bend90_factory,
            width_center=w,
            label=lbl,
            layer=layer,
            layers_cladding=layers_cladding,
        )
        for w, lbl in zip(widths, labels)
    ]
    y = -targets[0].size_info.height / 2 - spacing_v
    dx = targets[0].size_info.width + spacing_v
    x = dx / 2

    for d in targets:
        c.add_ref(d)
        d.movex(x)
        d.movey(y)
        x += dx

    c.rotate(90)
    c.move(c.size_info.sw, (0, 0))
    return c


def _TRCH_DASH_ISO(length=20.0, width=0.5, n=3, separation=2.0, label=""):
    c = pp.Component()
    _r = rectangle(x=width, y=length, layer=LAYER.WG)
    for i in range(n):
        r_ref = c.add_ref(_r)
        r_ref.movey(i * (length + separation))
        c.absorb(r_ref)

    if label:
        marker_label = manhattan_text(text=label, size=0.4, layer=LAYER.WG)
        _marker_label = c.add_ref(marker_label)
        _marker_label.movey((n - 1) * (length + separation + 4.0) + length / 2)
        c.absorb(_marker_label)

    return c


def _TRCH_DASH_DUO(
    length=20.0, gap=2.0, width=0.5, separation=4.0, n=3, x_offset=0.0, label=""
):

    _trench = _TRCH_DASH_ISO(length=length, width=width, n=n, separation=separation)
    dx = x_offset
    dy = gap + width
    c = pp.Component()
    t1 = c.add_ref(_trench)
    t2 = c.add_ref(_trench)
    t2.move((dy, dx))
    c.absorb(t1)
    c.absorb(t2)

    if label:
        marker_label = manhattan_text(text=label, size=0.4, layer=LAYER.WG)
        _marker_label = c.add_ref(marker_label)
        _marker_label.movey((n - 1) * (length + separation + 4.0) + length / 2)
        c.absorb(_marker_label)

    return c


LABEL_ITERATORS = {}


def gen_label_iterator(prefix=""):
    if not prefix:
        return ""

    if prefix and prefix not in LABEL_ITERATORS:
        LABEL_ITERATORS[prefix] = LabelIterator(prefix)
    return LABEL_ITERATORS[prefix]


class LabelIterator:
    def __init__(self, prefix):
        self.prefix = prefix
        self.counter = it.count(start=1)

    def __next__(self):
        return "{}{}".format(self.prefix, next(self.counter))

    def __iter__(self):
        return self


@pp.autoname
def TRCH_ISO(length=20.0, width=0.5):
    c = pp.Component()
    _r = c.add_ref(rectangle(x=width, y=length, layer=LAYER.SLAB150))
    c.absorb(_r)

    lblit = gen_label_iterator("TA")
    label = next(lblit)
    marker_label = manhattan_text(text=label, size=0.4, layer=LAYER.WG)
    _marker_label = c.add_ref(marker_label)
    _marker_label.movey(length / 2 + 4.0)
    c.absorb(_marker_label)

    return c


@pp.autoname
def TRCH_ISO_DL0(width=0.5, separation=2.0):
    lblit = gen_label_iterator("TB")
    return _TRCH_DASH_ISO(
        width=width, separation=separation, length=10.0, n=5, label=next(lblit)
    )


@pp.autoname
def TRCH_ISO_L20(width=0.5, separation=2.0):
    lblit = gen_label_iterator("TC")
    return _TRCH_DASH_ISO(
        width=width, separation=separation, length=20.0, n=3, label=next(lblit)
    )


@pp.autoname
def TRCH_DUO_DL0(width=0.5, separation=2.0, gap=3.0):
    lblit = gen_label_iterator("TD")
    return _TRCH_DASH_DUO(
        width=width, separation=separation, gap=gap, length=10.0, n=5, label=next(lblit)
    )


@pp.autoname
def TRCH_DUO_L20(width=0.5, separation=2.0, gap=3.0):
    lblit = gen_label_iterator("TE")
    return _TRCH_DASH_DUO(
        width=width, separation=separation, gap=gap, length=20.0, n=3, label=next(lblit)
    )


@pp.autoname
def TRCH_STG(width=0.5, separation=2.0, gap=3.0, n=6, length=20.0):
    lblit = gen_label_iterator("TF")
    return _TRCH_DASH_DUO(
        width=width,
        separation=separation,
        gap=gap,
        length=length,
        n=n,
        x_offset=length / 2 + separation / 2,
        label=next(lblit),
    )


if __name__ == "__main__":
    # c = cdsem_straight()
    # c = cdsem_straight_all()
    c = cdsem_uturn()
    c = cdsem_straight_density()
    # c = opcm()
    pp.show(c)



import pytest
import pp
from pp.import_gds import import_gds, add_ports_from_markers_center
from pp.port import auto_rename_ports

gdspaths = [pp.CONFIG["gdsdir"] / name for name in ["mmi1x2.gds", "mzi2x2.gds"]]


@pytest.mark.parametrize("gdspath", gdspaths)
def test_components_ports(gdspath, num_regression):
    c = import_gds(gdspath)
    add_ports_from_markers_center(c)
    auto_rename_ports(c)
    num_regression.check(c.get_ports_array())



"""
a container creates a new component that contains the original component inside with some extra elements:

it makes sure that some of the important settings are copied from the original component to the new component

"""

from typing import Callable
import functools
from inspect import signature
import pp


def container(component_function: Callable) -> Callable:
    """decorator for creating a new component that copies properties from the original component

    - polarization
    - wavelength
    - test_protocol
    - data_analysis_protocol

    Functions decorated with container will return a new component

    .. plot::
      :include-source:

      import pp

      @pp.container
      def add_padding(component, suffix='p', layer=pp.LAYER.M1):
          c = pp.Component(name=f"{component.name}_{suffix}")
          c << component
          w, h = component.xsize, component.ysize
          points = [[w, h], [w, 0], [0, 0], [0, h]]
          c.add_polygon(points, layer=layer)
          return c

      c = pp.c.waveguide()
      cp = add_padding(c)
      pp.plotgds(c)

    """

    @functools.wraps(component_function)
    def wrapper(*args, **kwargs):
        old = kwargs.get("component")
        if not old and args:
            old = args[0]
        if callable(old):
            old = old()
        if not isinstance(old, pp.Component):
            raise ValueError(
                f"container {component_function.__name__} requires a component, got `{old}`"
            )
        old = old or kwargs.get("component")
        new = component_function(*args, **kwargs)

        sig = signature(component_function)
        new.settings.update(**{p.name: p.default for p in sig.parameters.values()})
        new.settings.update(**kwargs)
        new.settings["component"] = old.settings.copy()
        new.settings["component_name"] = old.name
        new.settings["function_name"] = component_function.__name__
        new.test_protocol = new.test_protocol or old.test_protocol.copy()
        new.data_analysis_protocol = (
            new.data_analysis_protocol or old.data_analysis_protocol.copy()
        )
        new.wavelength = new.wavelength or old.wavelength
        new.polarization = new.polarization or old.polarization
        new.settings.pop("kwargs", "")
        return new

    return wrapper


@container
def add_padding(component, x=50, y=50, layers=[pp.LAYER.PADDING], suffix="p"):
    """ adds padding layers to component"""
    c = pp.Component(name=f"{component.name}_{suffix}")
    c << component
    points = [
        [c.xmin - x, c.ymin - y],
        [c.xmax + x, c.ymin - y],
        [c.xmax + x, c.ymax + y],
        [c.xmin - x, c.ymax + y],
    ]
    for layer in layers:
        c.add_polygon(points, layer=layer)
    return c


def test_container():
    old = pp.c.waveguide()
    suffix = "p"
    name = f"{old.name}_{suffix}"
    new = add_padding(component=old, suffix=suffix)
    assert new != old, f"new component {new} should be different from {old}"
    assert new.name == name, f"new name {new.name} should be {name}"
    # assert len(new.ports) == len(
    #     old.ports
    # ), f"new component {len(new.ports)} ports should match original {len(old.ports)} ports"
    # assert len(new.settings) == len(
    #     old.settings
    # ), f"new component {new.settings} settings should match original {old.settings} settings"
    return new


def test_container2():
    old = pp.c.waveguide()
    suffix = "p"
    name = f"{old.name}_{suffix}"
    new = add_padding(old, suffix=suffix)
    assert new != old, f"new component {new} should be different from {old}"
    assert new.name == name, f"new name {new.name} should be {name}"
    # assert len(new.ports) == len(
    #     old.ports
    # ), f"new component {len(new.ports)} ports should match original {len(old.ports)} ports"
    return new


def test_container_error():
    import pytest

    old = pp.c.waveguide()
    with pytest.raises(ValueError):
        add_padding(component2=old)  # will raise an error
    return old


if __name__ == "__main__":
    c = test_container2()
    pp.show(c)
